"""Simple database to track task execution.
"""

import collections
import doctest
import logging
import os
import datetime
import json
import sqlite3
import redis

from ox_herd import settings as ox_settings


def create(run_db=None):
    "Create and return RunDB reference based on run_db input."

    run_db = run_db if run_db else ox_settings.RUN_DB
    if run_db[0] == 'redis':
        return RedisRunDB()
    if run_db[0] == 'sqlite':
        return SqliteRunDB(run_db[1])

    raise ValueError('Could not understand run_db %s' % str(run_db))


class RunDB(object):
    """Abstract specification for database to track running of tasks.
    """

    def record_task_start(self, task_name, template=None):
        """Record that we are starting task with given name in database.

        :arg task_name:        String name for task.

        :arg template:    String indicating template to use in displaying
                          task result or None to use default.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :returns:   Task id to use in referring to task later (e.g, in
                    record_task_finish method). This is backend dependeant
                    and may be an integer or string or something else
                    depending on what is easiest for the backend.

        PURPOSE:    Record that we started something.

        """
        raise NotImplementedError

    def record_task_finish(self, task_id, return_value, status='finished',
                           json_blob=None, pickle_blob=None):
        """Record we finished a task.

        :arg task_id:        ID for task as returned by record_task_start.

        :arg return_value:  String return value of task.

        :arg status='finished':   String status of task.

        :arg json_blob=None:  Optional string representing json encoding of
                              task output. Using JSON to store the result
                              for later inspection is more portable.

        :arg pickle_blob=None:  Optional string representing python pickle
                                encoding of task output. Using JSON to store
                                the result for later inspection is more
                                portable, but you can use pickle if necessary.

        """
        raise NotImplementedError

    def delete_task(self, task_id):
        """Delete the task from the database.

        :arg task_id:        ID for task as returned by record_task_start.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:   Delete the task.

        """
        raise NotImplementedError

    def get_tasks(self, status='finished', start_utc=None, end_utc=None,
                  max_count=None):
        """Return list of TaskInfo objects.

        :arg status='finished':   Status of tasks to search. Should be one
                                  of entries from get_allowed_status().

        :arg start_utc=None: String specifying minimum task_start_utc.

        :arg end_utc=None:   String specifying maximum task_end_utc

        :arg max_count=None: Optional integer for max items to return

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :returns:       List of TaskInfo objects.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:        Main way to get information about the tasks run.

        """
        raw_tasks = self._help_get_tasks(status, start_utc, end_utc)
        return self.limit_task_count(raw_tasks, max_count)

    def _help_get_tasks(self, status='finished', start_utc=None, end_utc=None):
        """Return list of TaskInfo objects.

        :arg status='finished':   Status of tasks to search. Should be one
                                  of entries from get_allowed_status().

        :arg start_utc=None: String specifying minimum task_start_utc.

        :arg end_utc=None:   String specifying maximum task_end_utc

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :returns:       List of TaskInfo objects.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:        Main way to get information about the tasks run.
                        Intended to be called by get_tasks.

        """
        raise NotImplementedError

    @staticmethod
    def get_allowed_status():
        """Return list of allowed status strings for tasks.

        It is important to only use values from the allowed list so
        we can store effectively on things like redis.
        """
        return ['started', 'finished']

    @staticmethod
    def limit_task_count(task_list, max_count=-1):
        """Take list of TaskInfo items and returnt he last max_count items.
        """
        if max_count is None or max_count < 0 or len(task_list) < max_count:
            return task_list
        sorted_tl = list(sorted(
            task_list, key=lambda item: (
                item.task_end_utc if item.task_end_utc else
                item.task_start_utc)))
        return sorted_tl[-max_count:]

    def get_latest(self, task_name):
        """Return task_info for most recent finished task with given task_name.
        """
        raise NotImplementedError


class TaskInfo(object):
    """Python class to represent task info stored in database.
    """

    def __init__(  # pylint: disable=too-many-arguments
            self, task_id, task_name, task_start_utc=None,
            task_status=None, task_end_utc=None, return_value=None,
            json_data=None, pickle_data=None, template=None):
        self.task_id = task_id
        self.task_name = task_name
        self.task_start_utc = task_start_utc
        self.task_status = task_status
        self.task_end_utc = task_end_utc
        self.template = template
        self.return_value = return_value
        self.json_data = json_data
        self.pickle_data = pickle_data

    def __repr__(self):
        args = ', '.join(['%s=%s' % (
            name, repr(value)) for name, value in self.to_dict().items()])
        return '%s(%s)' % (self.__class__.__name__, args)

    def to_dict(self):
        """Return self as a dict.
        """
        return collections.OrderedDict([
            (name, getattr(self, name, '')) for name in [
                'task_id', 'task_name', 'task_start_utc', 'task_status',
                'task_end_utc', 'return_value', 'json_data', 'pickle_data',
                'template']])

    def to_json(self):
        """Return json version of self.
        """
        return json.dumps(self.to_dict())

    def run_time(self, round_to=2):
        """Return total running time if possible (-1 if task not finished)
        """
        if not self.task_end_utc:
            return -1
        result = 'UNKNOWN'
        try:
            fmt = '%Y-%m-%d %H:%M:%S.%f'
            end_utc_dt = datetime.datetime.strptime(
                self.task_end_utc, fmt)
            start_utc_dt = datetime.datetime.strptime(
                self.task_start_utc, fmt)
            result = (end_utc_dt - start_utc_dt).total_seconds()
            result = round(result, round_to)
        except Exception as problem:
            logging.exception('Could not parse start/end time of %s: %s',
                              self, problem)
            result = 'E:%s' % str(problem)
        return result


class RedisRunDB(RunDB):
    """Implementation of RunDB with redis backend.
    """

    def __init__(self):
        self.conn = redis.StrictRedis()
        self.my_prefix = ox_settings.REDIS_PREFIX + ':__'
        self.id_counter = self.my_prefix + 'task_id_counter'
        self.task_master = self.my_prefix + 'task_master' + '::'

    def delete_all(self, really=False):
        """Delete everything related to this from Redis.

        Only works if really=True.
        Mainly for testing; be *VERY* careful with this.
        """
        if not really:
            raise ValueError('Not doing delete_all since really=%s' % str(
                really))
        my_keys = list(self.conn.scan_iter(match=self.my_prefix + '*'))
        if my_keys:
            #  names = ' '.join([item.decode('utf8') for item in my_keys])
            self.conn.delete(*my_keys)

    def record_task_start(self, task_name, template=None):
        'Implement record_task_start for this backend.'

        if not task_name:
            raise ValueError('Must have non-empty task_name not "%s"' % str(
                task_name))
        if task_name[0:2] == ':_':
            raise ValueError('Invalid task name %s; cannot start with ":_"' % (
                str(task_name)))
        task_id = '%s_%s' % (task_name, datetime.datetime.utcnow().timestamp())
        task_key = self.task_master + task_id
        if self.conn.get(task_key):
            raise ValueError(
                'Cannot add task %s as %s since already exists' % (
                    str(task_name), task_id))
        info = TaskInfo(
            task_id, task_name, str(datetime.datetime.utcnow()),
            'started', template=template).to_json()
        add_result = self.conn.setex(
            task_key, ox_settings.OX_TASK_TTL, info)
        assert add_result, 'Got add_result = %s for %s; race condition?' % (
            add_result, task_id)

        return task_id

    def delete_task(self, task_id):
        """Delete desired id.
        """
        task_key = self.task_master + task_id
        self.conn.delete(task_key)

    def get_task_info(self, task_id):
        """Return dict representation of task with given task_id or None.
        """
        task_info = None
        task_key = self.task_master + task_id
        task_info_json = self.conn.get(task_key)
        if task_info_json:
            task_info = json.loads(task_info_json.decode('utf-8'))
        return task_info

    def get_task(self, task_id):
        """Get the task with the given task_id and return it as TaskInfo.
        """
        task_info = self.get_task_info(task_id)
        if task_info:
            return TaskInfo(**task_info)

        return None

    def record_task_finish(self, task_id, return_value, status='finished',
                           json_blob=None, pickle_blob=None):
        'Implement record_task_finish for this backend.'

        task_info = self.get_task_info(task_id)
        if not task_info:
            logging.error('Unable to update existing task with finish stats')
            logging.error('Will create finished but unstarted task')
            task_info = {'task_name': 'unknown', 'task_status': 'unknown'}

        if task_info['task_status'] == 'finished':
            raise ValueError('Cannot record_task_finish for %s; already ended.'
                             % str(task_info))
        task_info['task_end_utc'] = str(datetime.datetime.utcnow())
        task_info['return_value'] = return_value
        task_info['task_status'] = 'finished'
        task_info['json_data'] = json_blob
        task_info['pickle_data'] = pickle_blob
        task_key = self.task_master + task_id
        self.conn.setex(task_key,
                        ox_settings.OX_TASK_TTL, json.dumps(task_info))

    def _help_get_tasks(self, status='finished', start_utc=None, end_utc=None):
        """Return list of TaskInfo objects.

        :arg status='finished':   Status of tasks to search. Should be one
                                  entries from get_allowed_status().

        :arg start_utc=None: String specifying minimum task_start_utc.

        :arg end_utc=None:   String specifying maximum task_end_utc

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :returns:       List of TaskInfo objects.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:        Main way to get information about the tasks run.

        """
        result = []
        for key in self.conn.scan_iter(match=self.task_master + '*'):
            item_json = self.conn.get(key)
            item_kw = json.loads(item_json.decode('utf8'))
            if not (status is None or item_kw['task_status'] == status):
                continue
            if not (start_utc is None or item_kw.get(
                    'task_start_utc', start_utc) >= start_utc):
                continue
            if not (end_utc is None or item_kw.get(
                    'task_end_utc', end_utc) <= end_utc):
                continue
            result.append(TaskInfo(**item_kw))

        return result

    def get_latest(self, task_name):
        """Implementation of required get_latest method.

The redis implementation of this is not very efficient and could be
improved.
        """
        result = None
        my_tasks = self._help_get_tasks()
        for item in my_tasks:
            if (item.task_name != task_name or item.task_status != 'finished'):
                continue
            if result is None or (
                    item.task_end_utc > result.task_end_utc):
                result = item
        return result

    @staticmethod
    def _regr_test():
        """
>>> import os, tempfile, datetime, time, random, imp
>>> random_key = random.randint(0,10000000) # so tests do not collide
>>> print('Using random_key = %s' % str(random_key)) # doctest: +ELLIPSIS
Using random_key = ...
>>> from ox_herd.core import ox_run_db
>>> ox_run_db.ox_settings.REDIS_PREFIX += ('test_%s' % random_key)
>>> ox_run_db.ox_settings.OX_TASK_TTL = 10
>>> ignore = imp.reload(ox_run_db)
>>> db = ox_run_db.RedisRunDB()
>>> task_id = db.record_task_start('test')
>>> time.sleep(1)
>>> db.record_task_finish(task_id, 'test_return')
>>> t = db.get_tasks()
>>> len(t)
1
>>> t[0].task_name
'test'
>>> t[0].task_status
'finished'
>>> task_id = db.record_task_start('test_again')
>>> len(db.get_tasks('finished'))
1
>>> len(db.get_tasks(None))
2
>>> db.record_task_finish(task_id, 'test_again')
>>> t = db.get_tasks()
>>> len(t)
2
>>> max_list = db.get_tasks(max_count=1)
>>> len(max_list)
1
>>> max_list[0].task_name
'test_again'

Now verify that keys auto-expired in redis

>>> for i in range(10):
...     keys = db.conn.keys(ox_run_db.ox_settings.REDIS_PREFIX + '*')
...     if not keys:
...         break
...     logging.info('Sleeping a bit waiting for keys to expire: %s', keys)
...     time.sleep(2)
...
>>> db.conn.keys(ox_run_db.ox_settings.REDIS_PREFIX + '*')
[]

Now cleanup

>>> db.delete_all(really=True)
>>> db.conn.keys(ox_run_db.ox_settings.REDIS_PREFIX + '*')
[]

"""


class SqliteRunDB(RunDB):
    """Implementation of RunDB with sqlite backend.

    Redis is preferred, but SqliteRunDB is also possible with more
    configuration.
    """

    def __init__(self, db_path, allow_create=True):
        if not os.path.exists(db_path) and allow_create:
            logging.warning('No db file at %s; creating', str(db_path))
            self.create(db_path)
        self.conn = sqlite3.connect(db_path)

    @staticmethod
    def sql_to_create_tables():
        "Return SQL to create required database tables."

        sql = """CREATE TABLE task_info (
          task_id INTEGER PRIMARY KEY ASC,
          task_name text,
          task_start_utc text,
          task_status text,
          task_end_utc text,
          return_value text,
          json_blob text,
          pickle_blob text,
          template text
        );
        """
        return sql

    def create(self, db_path):
        "Create database at given path."

        sql = self.sql_to_create_tables()
        conn = sqlite3.connect(db_path)
        cursor = conn.cursor()
        cursor.execute(sql)
        conn.commit()
        conn.close()

    def record_task_start(self, task_name, template=None):
        'Implement record_task_start for this backend.'

        sql = '''INSERT INTO task_info (
          task_name, task_start_utc, task_status, template) VALUES (?, ?, ?, ?)
        '''
        cursor = self.conn.cursor()
        cursor.execute(sql, [task_name, datetime.datetime.utcnow(), 'started',
                             template])
        task_id = cursor.lastrowid
        self.conn.commit()
        assert task_id is not None, (
            'Expected 1 task id for insert but got %s' % str(task_id))
        return task_id

    def delete_task(self, task_id):
        """Delete desired id.
        """
        sql = '''DELETE FROM task_info WHERE task_id = ?'''
        self.conn.execute(sql, task_id)

    def record_task_finish(self, task_id, return_value, status='finished',
                           json_blob=None, pickle_blob=None):
        'Implement record_task_finish for this backend.'

        sql = '''UPDATE task_info
        SET task_end_utc=?, return_value=?, task_status=?,
            json_blob=?, pickle_blob=?
        WHERE task_id=?'''
        cursor = self.conn.cursor()
        utcnow = datetime.datetime.utcnow()
        cursor.execute(sql, [utcnow, return_value, str(status), json_blob,
                             pickle_blob, task_id])
        rowcount = cursor.rowcount
        if rowcount > 1:
            raise ValueError(
                'Impossible: updated multiple rows with single task_id %s' % (
                    str(task_id)))
        elif not rowcount:
            logging.error('Unable to update existing task with finish stats')
            logging.error('Will create finished but unstarted task')
            sql = '''INSERT INTO task_info (
              task_name, task_start_utc,
              task_id, task_end_utc, return_value, task_status
              json_blob=?, pickle_blob=?) VALUES (
              'unknown', 'unknown', ?, ?, ?, ?)'''
            cursor.execute(sql, [json_blob, pickle_blob, task_id,
                                 utcnow, return_value, status])

        self.conn.commit()

    def _help_get_tasks(self, status='finished', start_utc=None, end_utc=None):
        """Return list of TaskInfo objects.

        :arg status='finished':   Status of tasks to search. Should be one
                                  entries from get_allowed_status().

        :arg start_utc=None: String specifying minimum task_start_utc.

        :arg end_utc=None:   String specifying maximum task_end_utc

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :returns:       List of TaskInfo objects.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:        Main way to get information about the tasks run.

        """
        cursor = self.conn.cursor()
        sql = ['select * from task_info where task_status like ?']
        args = [status]
        if start_utc is not None:
            sql.append(' AND task_start_utc >= ?')
            args.append(str(start_utc))
        if end_utc is not None:
            sql.append(' AND (task_end_utc IS NULL OR task_end_utc >= ?)')
            args.append(str(end_utc))

        cursor.execute('\n'.join(sql), args)

        return [TaskInfo(*item) for item in cursor.fetchall()]

    @staticmethod
    def _regr_test():
        """
>>> import os, tempfile, datetime, time
>>> from ox_herd.core import ox_run_db
>>> db_file = tempfile.mktemp(suffix='.sql')
>>> db = ox_run_db.SqliteRunDB(db_file)
>>> task_id = db.record_task_start('test')
>>> time.sleep(1)
>>> db.record_task_finish(task_id, 'test_return')
>>> db.conn.close()
>>> del db
>>> os.remove(db_file)
>>> assert not os.path.exists(db_file)

"""


if __name__ == '__main__':
    doctest.testmod()
    print('Finished tests')



"""Helpers for flask views for ox_herd
"""

from functools import wraps

from flask import abort
from flask_login import current_user


def simple_role_check(*roles):
    """Decorator which specifies that a user must have all the specified roles.

    This is a simple hack to allow functioning if flask_security is
    not installed. If you want to do anything serious, you should
    install flask_security.
    """
    def wrapper(func):
        "Wrap view in role check"
        @wraps(func)
        def decorated_view(*args, **kwargs):
            """Decorate view with simple role check
            """
            if not roles:
                raise ValueError('Refusing to fake role check with no roles')
            user_roles = getattr(current_user, 'roles', [])
            if not set(user_roles).intersection(roles):
                abort(403, description=(
                    'User %s (roles=%s) lacks accpetable role: %s' % (
                        getattr(current_user, 'username', getattr(
                            current_user, 'id', 'unknown')),
                        str(user_roles), str(roles))))
            return func(*args, **kwargs)
        return decorated_view
    return wrapper



"""Forms for ox_herd commands.
"""

from wtforms import StringField
from ox_herd.core.plugins import base


class BackupForm(base.GenericOxForm):
    """Use this form to enter parameters for a new backup job.
    """

    bucket_name = StringField(
        'bucket_name', [], description=(
            'Name of AWS bucket to put backup into.'))

    bucket_name = StringField(
        'prefix', [], default='misc', description=(
            'Prefix to use in creating remote backup name.'))
    



"""Stub module for login to allow stand-alone usage.

If you import this module via something like

        from ox_herd.core import login_stub
        app.register_blueprint(login_stub.LOGIN_STUB_BP)

you will enable a simple stub login.
"""

from passlib.apps import custom_app_context as pwd_context

from flask import (Response, redirect, url_for, request,
                   Blueprint, flash, get_flashed_messages, escape)
from flask_login import LoginManager
from flask_login import (UserMixin, login_user, logout_user, login_required)

from ox_herd import settings

LOGIN_STUB_BP = Blueprint('login_stub', __name__)
LOGIN_MANAGER = LoginManager()

@LOGIN_STUB_BP.record_once
def on_load(state):
    "Initialize login manager"

    LOGIN_MANAGER.init_app(state.app)

class User(UserMixin):

    _userdict = {}

    def __init__(self, uname, roles=None):
        self.id = uname
        self.name = uname
        self.roles = roles if roles else []
        self.__class__._userdict[self.id] = self

    @classmethod
    def load(cls, uid):
        "Load user with given ID."
        return cls._userdict[uid]


@LOGIN_STUB_BP.route("/login", methods=["GET", "POST"])
def login():
    if request.method == 'POST':
        username = request.form['username']
        password = request.form['password']
        password_hash = settings.STUB_USER_DB.get(username, 'disabled')
        if password_hash != 'disabled' and pwd_context.verify(
                password, password_hash):
            user = User(username, settings.STUB_USER_ROLES.get(username, []))
            login_user(user)
            next_url = request.args.get("next", '')
            next_url = next_url if next_url.strip().lower() != 'none' else (
                url_for('ox_herd.index'))
            return redirect(next_url)
        else:
            flash('Login failed to login as %s; try again' % username)
            if password_hash == 'disabled':
                flash('WARNING: username %s is disabled' % username)
            return redirect(url_for('login_stub.login'))
    else:
        messages = get_flashed_messages()
        flashes = '<UL>\n%s\n</UL>' % '\n'.join([
            '<LI>%s</LI>' % escape(m) for m in messages]) if messages else ''
        return Response(
            flashes + '''
            <h1>ox_herd login</h1>
            <p>
            Welcome to ox_herd! 
            You are running the stand-alone ox_herd server.            
            </p>
            <p>
            To interact, you need to login.
            See the ox_herd/settings.py file to setup a trivial 
            username/password dictionary to use with this stub login system.
            <p>
            Ideally, we will develop a better login setup for ox_herd or you
            can use ox_herd as a blueprint in a larger flask fraemwork that
            already handles its own login.
            </p>
        <form action="" method="post">
            <p><input type=text name=username>
            <p><input type=password name=password>
            <p><input type=submit value=Login>
        </form>
        ''')


# somewhere to logout
@LOGIN_STUB_BP.route("/logout")
@login_required
def logout():
    "Simple logout function."
    logout_user()
    return Response('<p>Logged out</p>')


@LOGIN_MANAGER.user_loader
def load_user(userid):
    "Callback to reload the user object"

    return User.load(userid)



@LOGIN_MANAGER.unauthorized_handler
def unauthorized():
    """Return response for when user is not logged in.
    """
    return redirect(url_for('login_stub.login'))



"""Basic tests.
"""

import logging
import random

import requests

from ox_herd.core.utils import test_utils


class SimpleTest(test_utils.SelfContainedTest):
    """Run some simple tests.
    """

    _user_info = {}

    @classmethod
    def setUpClass(cls):  # pylint: disable=invalid-name
        if not cls._user_info:
            cls._user_info = {
                'generic': str(random.randint(0, 1e20)),
                'test_admin': str(random.randint(0, 1e20))
                }
        stub_info = ','.join(['%s:%s' % (k, v)
                              for k, v in cls._user_info.items()])
        cls._serverInfo = test_utils.start_server(
            stub_user=stub_info, stub_roles='test_admin:admin')
    
    def test_good_health_check(self):
        "Verify that health check is OK"

        logging.debug('Doing test_good_health_check on %s', str(self))
        result = requests.get(
            'http://localhost:%i/ox_herd/health_check?token=%s' % (
                self._serverInfo.port, self._serverInfo.health_token))
        self.assertEqual(result.status_code, 200)

    def test_bad_health_check(self):
        "Verify that health check returns 500 if try on non-existing queue"

        logging.debug('Doing test_bad_health_check on %s', str(self))
        queue = 'some_random_bad_queue_%s' % random.randint(0, 1e20)
        params = 'token=%s&check_queues=%s' % (
            self._serverInfo.health_token, queue)
        expect_bad = requests.get(
            'http://localhost:%i/ox_herd/health_check?%s' % (
                self._serverInfo.port, params))
        self.assertEqual(expect_bad.status_code, 500)

    def test_configure_job(self):
        """Test that we can trigger job configuration without error.
        """
        session = requests.session()
        session.post('http://localhost:%i/login' % (
            self._serverInfo.port), {
                'username': 'test_admin',
                'password': self._user_info['test_admin']})
        url = 'http://localhost:%i/ox_herd/use_plugin' % self._serverInfo.port
        result = session.get(url, params={
            'plugname': 'ox_herd.core.plugins.example_psutil_plugin',
            'plugcomp': 'CheckCPU'})
        self.assertEqual(result.reason, 'OK')
        self.assertEqual(result.status_code, 200)

    def test_access_restricted_ox_herd(self):
        "Verify that non-admin user cannot access ox_herd views"

        self.check_access_restricted_ox_herd(
            'generic', 'FORBIDDEN', 403)
        self.check_access_restricted_ox_herd('test_admin')

    def check_access_restricted_ox_herd(
            self, username, reason='OK', status=200):
        session = requests.session()
        result = session.post('http://localhost:%i/login' % (
            self._serverInfo.port), {'username': username,
                                     'password': self._user_info[username]})
        self.assertEqual(result.status_code, 200)

        url = 'http://localhost:%i/ox_herd/show_scheduled' % (
            self._serverInfo.port)
        result = session.get(url)
        self.assertEqual(result.reason, reason)
        self.assertEqual(result.status_code, status)



"""Flask Blueprnit for ox_herd.

This module provides the OX_HERD_BP variable representing the main
ox_herd blueprint. If you are incorporating ox_herd into your own
flask application, you should call register_blueprint(OX_HERD_BP)
appropriately.
"""

import logging
from flask import Blueprint
from ox_herd.core.plugins import manager as plugin_manager

class OxHerdBlueprint(Blueprint):
    """Subclass flask Blueprint to provide custom blueprint for ox_herd.

    Mainly this blueprint overrides register to make sure to start plugins.
    """
    
    def register(self, app, *args, **kwargs):
        """Override default register method to also activate plugins.
        
        :arg app, *args, **kwargs:  As for usual Blueprint.register method.
        
        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
        
        :returns:  As usual for Blueprint.register.
        
        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
        
        PURPOSE:   Override registration so we can start plugins.
        
        """
        result = Blueprint.register(self, app, *args, **kwargs)
        self.start_oh_plugins(app)
        return result
        
    def start_oh_plugins(self, app):
        """Start ox_herd plugins. Meant to be called by register method.
        
        :arg app:    Flask app we are using to regsiter blueprints.
        
        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-
        
        PURPOSE:   This method finds active plugins, registers any blueprints
                   those plugins provide, and generally gets the plugin
                   manager setup.
        """
        manager = plugin_manager.PluginManager()
        manager.activate_plugins()
        actives = manager.get_active_plugins()
        logging.warning('Activating plugins: %s', str(actives))
        for pname, plug in actives.items():
            bprint = plug.get_flask_blueprint()
            if bprint and not isinstance(bprint, Blueprint):
                raise TypeError(
                    'Plugin %s.get_flask_blueprint gave %s not Blueprint' % (
                        pname, str(bprint)))
            if bprint:
                app.register_blueprint(bprint)
            else:
                logging.debug('No blueprint to register for plugin %s',
                              pname)
        

OX_HERD_BP = OxHerdBlueprint('ox_herd', __name__, template_folder='templates',
                             static_folder='static', url_prefix='/ox_herd')




"""Module for managing ox_herd plugins.
"""

import os
import logging
import inspect
import importlib
from ox_herd import settings
from ox_herd.core.plugins import base


class PluginManager(object):

    __active_plugins = {}

    @classmethod
    def activate_plugins(cls):
        """Activate plugins.
        """
        active_names = list(settings.OX_PLUGINS)
        active_set = set(active_names)
        env_plugs = os.getenv('OX_PLUGINS', '').split(':')
        for name in env_plugs:
            if name and name not in active_set:
                active_names.append(name)
        for name in active_names:
            if name in cls.__active_plugins:
                logging.warning('Plugin %s already activated; skipping', name)
                continue
            logging.info('Importing %s', name)
            my_mod = importlib.import_module(name)
            plug = cls.make_plugin_from_module(name, my_mod)
            assert isinstance(plug, base.OxPlugin)
            cls.__active_plugins[name] = plug

    @classmethod
    def get_active_plugins(cls):
        return dict(cls.__active_plugins)

    @classmethod
    def make_plugin_from_module(cls, name, my_mod):
        maker = getattr(my_mod, 'get_ox_plugin', None)
        if maker:
            return maker()
        logging.debug('No get_ox_plugin function for plugin %s; %s.',
                      name, 'Using default plugin maker.')
        return cls.default_plugin_maker(name, my_mod)

    @classmethod
    def default_plugin_maker(cls, name, my_mod):
        plugins = []
        klasses = []
        components = []
        for dname in dir(my_mod):
            item = getattr(my_mod, dname)
            if isinstance(item, base.OxPlugin):
                plugins.append((dname, item))
            elif inspect.isclass(item):
                if issubclass(item, base.OxPlugin):
                    klasses.append((dname, item))
                elif issubclass(item, base.OxPluginComponent):
                    components.append((dname, item))
        if plugins:
            if len(plugins) > 1:
                msg = "Can't make default plugin for %s because:\n%s." % (
                    name, ('Found %i plugin candidates: %s' % (
                        len(plugins), [pair[0] for pair in plugins])))
                raise ValueError(msg)
            else:
                return plugins[0][1]
        if klasses:
            if len(klasses) > 1:
                msg = "Can't make default plugin for %s because:\n%s." % (
                    name, ('Found %i plugin candidates classes: %s' % (
                        len(klasses), [pair[0] for pair in klasses])))
                raise ValueError(msg)
            else:
                instances = [c[1](name=c[0]) for c in components]
                return klasses[0][1](components=instances)
        if components:
            instances = [c[1](name=c[0]) for c in components]
            my_plugin = base.TrivialOxPlugin(instances, name, (
                'Automatically created plugin for %s.' % name))
            return my_plugin
            
        msg = 'Could not find any plugins or components in %s' % str(name)
        logging.error('%s\nSearched:\n%s\n', msg, str(list(dir(my_mod))))
        raise ValueError(msg)
            



"""Simple example of an ox_herd plugin.

This is module serves as an example of how to create an ox_herd plugin.
It provides a single class: CheckCPU which uses the psutil module (which
you may have to install on your system if you don't have it already).

Provided that you include this plugin in the OX_PLUGINS variable in
ox_herd.settings or in your OX_PLUGINS environment variable, this will be
picked up automatically and used as a plugin.

In general, there are a variety of more complicated things you can do
in configuring plugins. See documentation on plugins or see the pytest_plugin
example for more details.
"""

import json
import psutil

from ox_herd.core.plugins import base


class CheckCPU(base.OxPlugTask):
    """Class to check and report CPU usage.

    This is meanly meant to serve as an example of a minimal plugin.
    All we do is implement the main_call method.
    """

    @classmethod
    def main_call(cls, ox_herd_task):
        """Main method to lookup CPU usage.

        :arg ox_herd_task:   Instance of a CheckCPU task perhaps containing
                        additional data (e.g., ox_herd_task.name). If your
                        main_call does not need arguments, you can basically
                        just ignore ox_herd_task. If you do want to be able
                        to pass in arguments, see a more detailed discussion
                        of how to get arguments from the user and configure
                        a task in the full plugin documentation.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :returns:       Dictionary with 'return_value' and 'json_blob' as
                        required for OxPluginComponent.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:        Use the psutil module to lookup CP usage.

        """
        cpu = psutil.cpu_percent()
        pid_func = getattr(psutil, 'get_pid_list', getattr(psutil, 'pids'))
        if pid_func is None:
            raise ValueError(
                'Do not know how to get pids with version %s of psutil' % (
                    getattr(psutil, '__version__', 'unknown')))
        pids = len(pid_func())
        data = {'cpu_percent': cpu, 'num_pids': pids}
        return {
            'return_value': 'Task %s completed succesfully: cpu=%s' % (
                ox_herd_task.name, cpu), 'json_blob': json.dumps(data)
            }



"""Tests for awstools plugin.
"""

import shutil
import os
import tempfile


from ox_herd.core.plugins.awstools_plugin import core


class TestableBackupTask(core.BackupPostgresToAWS):
    """Sub-class of BackupPostgresToAWS for testing.
    """

    @classmethod
    def make_dump_cmdline(cls, ox_herd_task, outfile):
        """Override to just cat file for dump to simplify testing.
        """
        return ['cp', cls.get_conn_string(ox_herd_task), outfile]

    @staticmethod
    def get_conn_string(ox_herd_task):
        """Provide simplified version for testing in case we want override.

Helpful if we deal with a sqlite connection string.
        """
        return ox_herd_task.conn_string.split('://')[-1]


def test_basic_operation():
    "Test simplified version of backup task."
    try:
        db_loc = tempfile.mktemp()
        backup_loc = tempfile.mktemp()
        os.mkdir(backup_loc)
        open(db_loc, 'w').write('test_data.txt')
        task = TestableBackupTask(
            name='test_task', conn_string=db_loc, prefix='test',
            bucket_name='@'+backup_loc)
        result = task.main_call(task)
        assert result['return_value'].split()[:4] == [
            'Finished', 'dump:', 'extra', 'messages="b\'\'".']
    finally:
        for name in [db_loc]:
            if os.path.exists(name):
                os.remove(name)
        if os.path.exists(backup_loc):
            shutil.rmtree(backup_loc)



"""Utilities used by other tests.
"""

import sys
import os
import socket
import logging
import subprocess
import weakref
import unittest
import random

from passlib.apps import custom_app_context as pwd_context

import ox_herd


class ServerInfo:  # pylint: disable=too-few-public-methods
    """Information about an server running for testing.
    """

    def __init__(self, server, port, health_token, stub_user=None):
        self.server = server
        self.port = port
        self.health_token = health_token
        self.stub_user = stub_user


def start_server(
        debug='1',     # Need debug 1 so we can use tests endpoint
        reloader='0',  # Need to use reload 0 to prevent flask weirdness
        ox_port=None, cwd=None, stub_user=None, stub_roles=None):
    """Start flask server.

    :param debug='1':     Whether to run in debug mode.

    :param reloader='0':  Whether to use reloader in server. Often this
                          causes problems so it is best not to.

    :param ox_port=None:  Optional port for server. If None, we choose.

    :param cwd=None:      Current working directory to run in. If None,
                          we run in ox_herd root.

    :param stub_user=None:  Optional --stub_user argument (in the form
                            of <user>:<passwd>) for stub db for testing.
                            If None, we will randomly generate it.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    :return:  Instance of ServerInfo containing information about the
              server we started.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:  Use python subprocess module to start an instance of the
              server in the background. You can use the returned
              ServerInfo instance to interact with the server. This is
              useful for making the test self-contained.

              You can call kill_server to cleanup.
    """
    health_token = random.randint(0, 1e20)
    server_script = os.path.join(os.path.dirname(ox_herd.__file__),
                                 'scripts', 'serve_ox_herd.py')
    cwd = cwd if cwd else os.path.join(os.path.dirname(ox_herd.__file__))
    ox_port = ox_port if ox_port else str(find_free_port())
    cmd = [sys.executable, server_script, '--debug', str(debug),
           '--port', ox_port, '--health_token', str(health_token),
           '--plugin', 'ox_herd.core.plugins.example_psutil_plugin']
    if not stub_user:
        stub_user = '%s:%s' % (
            random.randint(0, 1e10), random.randint(0, 1e10))
    if stub_roles:
        cmd.extend(['--stub_roles', stub_roles])
    cmd.extend(['--stub_user', stub_user, '--hash_stub', '0'])
    print(cmd)
    server = run_cmd(cmd=cmd, cwd=cwd, timeout=-4)
    return ServerInfo(server, int(ox_port), health_token,
                      stub_user=stub_user)


def run_cmd(cmd, cwd=None, timeout=30):
    """Helper to run a command in a subprocess.

    :param cmd:   List of strings for commands to run.

    :param cwd=None:    Optional current working directory string.
                        If None, we use ox_herd root.

    :param timeout=30:  Timeout in seconds before killing command.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    :return:  The subprocess created to run the command.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:  Provides some convenience for calling subprocess.Popen
              to run subprocess commands.

    """
    if cwd is None:
        cwd = os.path.join(os.path.dirname(ox_herd.__file__))
    logging.info('Running cmd: %s', str(cmd))
    proc = subprocess.Popen(cmd, cwd=cwd)
    if timeout is not None:
        try:
            result = proc.wait(timeout=abs(timeout))
        except subprocess.TimeoutExpired:
            if timeout < 0:
                result = proc.returncode  # timeout is OK
            else:
                raise  # timeout not ok so re-raise it
    else:
        result = proc.returncode
    if result:
        raise ValueError('Error code "%s" in cmd "%s"' % (
            result, proc))
    return proc


def find_free_port():
    "Find and return a free port number"

    result = None
    my_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        my_sock.bind(('', 0))
        my_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        result = my_sock.getsockname()[1]
    finally:
        my_sock.close()
    return result


def cleanup(cleanup_files, server_info=None):
    """Cleanup files and server.

    :param cleanup_files:  List of files to remove.

    :param server_info=None:  Instance of ServerInfo to clean.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    :return:  A weakref to server_info.server if we stop it.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:  Cleanup by remoging cleanup_files and stopping
              server_info.server if they exist.

    """
    for name in cleanup_files:
        if os.path.exists(name):
            logging.info('Removing file %s', name)
            os.remove(name)
            assert not os.path.exists(name), 'Could note remove ' + name

    if server_info is None:
        return None
    kill_server(server_info)
    dead_server = weakref.ref(
        server_info.server)  # keep weakref for debugging

    return dead_server


def kill_server(server_info):
    """Kill server to cleanup.

    :param server_info:   Instance of ServerInfo.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:  Terminate server in server_info.server.

    """
    timeout = 10
    server_info.server.terminate()
    status = server_info.server.wait(timeout=timeout)
    if status is None:
        logging.warning('Server failed to terminate; killing')
        server_info.server.kill()
        status = server_info.server.wait(timeout=timeout)
        if status is None:
            raise ValueError('Server refused to die')


class SelfContainedTest(unittest.TestCase):
    """Self contained test to start a flask server and tear it down.
    """

    _serverInfo = None

    @classmethod
    def setUpClass(cls):  # pylint: disable=invalid-name
        """Setup the class to have a running server.
        """
        if cls._serverInfo is not None:
            raise ValueError('Refusing to setup since _serverInfo not empty')
        cls._serverInfo = start_server()

    @classmethod
    def tearDownClass(cls):  # pylint: disable=invalid-name
        """Tear down the class.
        """
        kill_server(cls._serverInfo)
        cls._serverInfo = None



"""Core tools for ox_herd flask views.
"""

import logging

from ox_herd.ui.flask_web_ui.ox_herd import OX_HERD_BP
from ox_herd import settings
from ox_herd.core.utils import ox_decs
from ox_herd.ui.flask_web_ui.ox_herd import helpers


ROLES_CHECK = None

try:
    from flask_security import roles_accepted
    ROLES_CHECK = roles_accepted
except ImportError as problem:
    logging.warning('Error on importing roles_accepted: %s; will fake it',
                    str(problem))


if not ROLES_CHECK:
    ROLES_CHECK = helpers.simple_role_check


def ox_herd_route(*args, extra_roles=(), noauth=False, **kwargs):
    """Decorator to ensure user is allowed to access ox_herd views.

    This is intended to be called instead of OX_HERD_BP.route for when
    you want a route which forces authentication + ox_herd permissions.
    Use this and **NOT** OX_HERD_BP.route.
    """
    def make_decorator(my_func):
        """Make combined decorator for given input `my_func`.

        This is a little tricky because we want to make a decorator
        that takes in my_func, applies the OX_HERD_BP.route decorator and
        makes sure all the docs are fixed up right.

        This function can then be applied as a one shot decorator to do
        the above where as if you tried to do it separately then the
        docs would be hard to get right.
        """
        if noauth:
            return OX_HERD_BP.route(*args, **kwargs)(my_func)

        if settings.OX_HERD_ROLES:
            with_role = ROLES_CHECK(*(settings.OX_HERD_ROLES + list(
                extra_roles)))(my_func)
        else:  # Do not wrap with role check
            with_role = my_func
        with_route = OX_HERD_BP.route(*args, **kwargs)(with_role)
        result = with_route
        ox_decs.fix_doc(result, OX_HERD_BP.route)
        return result
    return make_decorator


def make_form_for_task(my_args):
    """Take an ox_herd plugin instance (or something like it) and make form.

This tries to create a form to configure an ox_herd task.
    """
    # Try to get the deprecated get_flask_form first.
    # Even if they are using the preferred get_flask_form_via_cls, the
    # get_flask_form should call that. But if they are using the old
    # get_flask_form then need to go through that flow first.
    maker = getattr(my_args, 'get_flask_form', None)
    if maker is None:
        logging.warning('could not find get_flask_form method of %s',
                        str(my_args))
        maker = getattr(my_args, 'get_flask_form_via_cls', None)
    if maker is None:
        msg = ("Item %s does not inherit from OxPlugin so can't configure.\n"
               "Rewrite that class to inherit from OxPlugin or at least\n"
               "provide a get_flask_form_via_cls method.")
        raise ValueError(msg)
    my_form_cls = maker()
    my_form = my_form_cls(obj=my_args)
    return my_form



"""Provides plugin containing ox_herd tools for AWS.
"""

from flask import Blueprint
from ox_herd.core.plugins.awstools_plugin.core import OxHerdAWSToolsPlugin


AWSTOOLS_BP = Blueprint(
    'awstools_plugin', __name__, template_folder='templates')
OxHerdAWSToolsPlugin.set_bp(AWSTOOLS_BP)


def get_ox_plugin():
    """Required function for module to provide plugin.
    """
    return OxHerdAWSToolsPlugin



"""Module to handle ox_herd tasks.

An OxHerdTask is a general class to represent a task managed by ox_herd.
Users are meant to sub-class OxHerdTask for their own use. See docstring for
OxHerdTask for details.

You can find some example sub-classes of OxHerdTask in the simple_ox_tasks.py
module.
1;5202;0c"""

import shlex
import copy
import logging
import json

from ox_herd import settings as ox_settings
from ox_herd.core import ox_run_db


class OxHerdTask:
    """Generic task class for ox_herd.

    Users should sub-class OxHerdTask and implement the main_call method.

    Due to the way python rq is setup, it's painful to try and pass class
    instances. Instead, it's better to have your task class have only
    class methods and pass the data in the kwargs part of a python rq job.
    This lets you more easily inspect the data that a task will be running on.
    To faciliate this, the OxHerdTask is setup to use only classmethods
    and pass an instance of itself in the kwargs of python rq.

    Basically, you create an INSTANCE of OxHerdTask (or sub-class) with the
    desired data to configure the task. You then pass that INSTANCE to
    the add_to_schedule method of OxScheduler which sets things up so that
    if you are using python rq, the INSTANCE is in kwargs and the function
    is a class method which uses those kwargs.

    The simple_ox_tasks has some examples.

    TL;DR = Use only classmethod in OxHerdTask (or sub-classes) and expect
            an instance to be passed in with data.
    """

    # Fields that affect how we pass job into python rq
    rq_fields = ['func', 'queue_name', 'timeout', 'cron_string']

    def __init__(self, name, func=None, run_db=None, queue_name=None,
                 timeout=None, cron_string=None):
        """Initializer.

        :arg name:     String name for the task.

        :arg func=None:  Function to run remotely. Usually this is the
                         run_ox_task classmethod of a sub-class of the
                         OxHerdTask and that is what we use if func is None.
                         The instance of OxHerdTask will be passed
                         as a keyword arg to func with key 'ox_herd_task', so
                         make sure func can be called as
                         func(ox_herd_task=task).

        :arg run_db=None:    Database used to track execution of the task. If
                             None is provided, we use default db.

        :arg queue_name=None:  Optional string representing which queue
                               to run. If not provided, then we use the
                               first item from ox_settings.QUEUE_NAMES.

        :arg timeout=None:     Optional timeout for job.

        :arg cron_string=None: Optional string in cron format saying how
                               the job should be scheduled.

        """
        self.name = name
        self.func = func if func is not None else self.run_ox_task
        self.run_db = run_db if run_db else self.choose_default_run_db()
        self.queue_name = queue_name if queue_name else shlex.split(
            ox_settings.QUEUE_NAMES)[0]
        self.timeout = timeout
        self.cron_string = cron_string
        self.rdb_job_id = None  # job_id inside our own RunDB

    @staticmethod
    def make_copy(ox_herd_task, name_suffix='_copy'):
        """Helper function to make a copy of the task.

        :arg ox_herd_task:        Instance of OxHerdTask to copy.

        :arg name_suffix='_copy': Optional suffix to append to name.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :returns:       Copied task.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:    We sometimes want to make copies of a task (e.g.,
                    if the user wants to launch a copy of a scheduled task).
                    This method makes a copy but with a different name. It
                    also allows sub-classes to control how copies are made.
        """
        args = copy.deepcopy(ox_herd_task)
        args.name += name_suffix
        return args

    @staticmethod
    def choose_default_run_db():
        """Chose default settings for run_db baed on settings for ox_herd.

        """

        run_db = ox_settings.RUN_DB
        if run_db[0] == 'redis':
            return run_db
        if run_db[0] == 'sqlite':
            if run_db[1]:
                return run_db
            raise ValueError('Need to provide valid path for sqlite run_db')

        raise ValueError('Could not understand run_db setting: %s'
                         % str(run_db))

    @staticmethod
    def get_template_name():
        """Return string for jinja template to use in display task result.

        By default we just display some information which every task should
        have available.

        If you register your own blueprints or otherwise get jinja templates
        into your path, you can override this to return your own templates.

        The task will be passed in as as task_data.
        """
        return 'generic_ox_task_result.html'

    @classmethod
    def main_call(cls, ox_herd_task):
        """Main function to run the task.

        Sub-classes should override to do something useful.

        The return value should either be a simple string of 80 characters
        or less describing the result or a dict.

        If a dict is returned, it should be as follows:

           { 'return_value' : <simple string or return code>,
             'json_blob' : <string as produced by json.dumps>,
             'pickle_blob' : <string as produced by pickle.dumps }

        You should use return_value for simple, small return codes like a
        'completed succesfully' message or 'invalid data' message or something.
        You should use json_blob for more complicated results when possible
        as that is most portable. If you have really complex return values,
        use pickle_blob. If provided, these values will be stored in the
        RunDB for later inspection by the web UI or other tools.
        """
        raise NotImplementedError

    @classmethod
    def pre_call(cls, ox_herd_task, rdb):
        """Will be called before main_call.

        :arg ox_herd_task:  The instance of the task to do pre_call for.

        :arg rdb:    Instance of RunDB to record job start and job end.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:   Run some operations before main_call.

        This does things like stores the fact that we started the job in
        a database. If users override, they should probably call this, or
        implement their own job tracking. The idea is that users can just
        use the default pre_call and only override main_call for their job.
        """
        assert ox_herd_task.rdb_job_id is None, (
            'Cannot have rdb_job_id set before callign pre_call.')
        ox_herd_task.rdb_job_id = rdb.record_task_start(
            ox_herd_task.name, ox_herd_task.get_template_name())

    @classmethod
    def post_call(cls, ox_herd_task, rdb, call_result, status='finished'):
        """Called just after main_call finishes.

        :arg ox_herd_task:  The instance of the task to do post_call for.

        :arg rdb:    Instance of RunDB to record job start and job end.

        :arg call_result:  Value returned by main_call (either a string or
                           dict); see docs for main_call return value.

        :arg status='finished':    Final status of job.


        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:        Tracks completion of the job.

        This does things like stores the fact that we finished the job in
        a database. If users override, they should probably call this, or
        implement their own job tracking.
        """
        rval = {}
        if isinstance(call_result, str):
            rval['return_value'] = call_result
        elif isinstance(call_result, dict):
            rval = dict(call_result)
        elif hasattr(call_result, 'to_dict'):
            as_dict = call_result.to_dict()
            rval = {'return_value': as_dict.pop('return_value')}
            rval['json_blob'] = json.dumps(as_dict)
        else:
            raise TypeError(
                'call_result from main_call not str/dict/to_dict; got %s' % (
                    str(call_result)))
        json_blob_dict = {}
        if 'json_blob' in rval:
            json_blob_dict = json.loads(rval.pop('json_blob'))
        for name in list(rval):
            if name not in ('task_id', 'return_value', 'status',
                            'pickle_blob', 'call_result'):
                json_blob_dict[name] = rval.pop(name)
        rval['json_blob'] = json.dumps(json_blob_dict)

        rdb.record_task_finish(ox_herd_task.rdb_job_id, status=status, **rval)

    @classmethod
    def run_ox_task(cls, ox_herd_task):
        """Entry point to run the task.

        This is what python rq or other managers will use to start the task.
        """
        job_name = ox_herd_task.name
        run_db = ox_herd_task.run_db
        rdb = ox_run_db.create(run_db)
        cls.pre_call(ox_herd_task, rdb)
        try:
            result = cls.main_call(ox_herd_task)
        except Exception as problem:
            logging.error('For job %s; storing exception result: %s',
                          job_name, str(problem))
            cls.post_call(ox_herd_task, rdb, str(problem), 'exception')
            raise
        cls.post_call(ox_herd_task, rdb, result)

        return result

    def get_display_fields(self, generics=(
            ('CRON string', 'cron_string'), ('timeout', 'timeout'),
            ('URL', 'url'), ('pytest args', 'pytest_cmd'))):
        """Get fields to display for task.

        :param generics: List of pairs of the form (TITLE, FIELD) where
                         TITLE is a string name and FIELD is a string
                         naming the property in self. We try to get the
                         values for these fields.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:  List of pairs of the form (TITLE, VALUE) where TITLE
                  is the title for the field and VALUE is the value for
                  the field.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Get fields to display in showing task configuration. By
                  default we look for various generic fields as in the
                  generics argument as well as looking for fields from
                  self.get_flask_form() if that is defiend and returns
                  something useful. Note that you should probably
                  implement get_flask_form_via_cls not get_flask_form since
                  get_flask_form will by default call get_flask_form_via_cls.

                  Sub-classes could override if desired.
        """
        result, found = [], {}
        generics = list(generics)
        my_form_cls_maker = getattr(self, 'get_flask_form_via_cls', None)
        if my_form_cls_maker:
            my_form_cls = my_form_cls_maker()  # pylint: disable=not-callable
            my_form = my_form_cls()
            generics.extend([(f, f) for f, _ in my_form.data.items()])
        for name, field in generics:
            if field not in found and hasattr(self, field):
                found[field] = True
                result.append((name, getattr(self, field)))

        return result



"""Module containing some plugin to run pytest.
"""

import os
import shutil
import datetime
import gzip
import logging
import tempfile
import subprocess

import boto3

from ox_herd.core.plugins import base
from ox_herd.core.ox_tasks import OxHerdTask
from ox_herd.core.plugins.awstools_plugin import forms


class OxHerdAWSToolsPlugin(base.OxPlugin):
    """Plugin to provide AWS services for ox_herd
    """

    __blueprint = None

    @classmethod
    def set_bp(cls, my_bp):
        """Set blueprint class variable.
        """
        cls.__blueprint = my_bp

    def get_flask_blueprint(self):
        """Implement as required by OxPlugin."""

        return self.__class__.get_bp()

    @classmethod
    def get_bp(cls):
        """Get blueprint class variable.
        """
        return cls.__blueprint

    def name(self):
        """Implement as required by OxPlugin."""

        return 'awstools_plugin'

    def description(self):
        """Implement as required by OxPlugin."""

        return "Plugin to provide AWS tools for ox_herd."

    def get_components(self):
        return [BackupPostgresToAWS()]


class BackupPostgresToAWS(OxHerdTask, base.OxPluginComponent):
    """Task to backup postgres instance to AWS
    """

    def __init__(self, *args, conn_string=None, prefix=None,
                 bucket_name=None, timeout=1800, **kwargs):
        """Initializer.

        :arg *args:    Argumnets to OxHerdTask.__init__.

        :arg conn_string:  Connection string to database.

        :arg **kwargs:     Keyword arguments to OxHerdTask.__init__.

        """
        OxHerdTask.__init__(self, *args, **kwargs)
        self.conn_string = conn_string
        self.prefix = prefix
        self.bucket_name = bucket_name
        self.timeout = timeout

    @classmethod
    def get_flask_form_via_cls(cls):
        """Get flask form class to enter parameters.
        """
        result = forms.BackupForm
        logging.debug('Providing form %s for cls %s', result, cls)
        return result

    @staticmethod
    def get_conn_string(ox_herd_task):
        """Get connection string from class.

By default we just look for ox_herd_task.conn_string but sub-classes can
override to pull from secret vault or something.
        """
        return ox_herd_task.conn_string

    @classmethod
    def main_call(cls, ox_herd_task):

        rval = cls.backup_pion_db(ox_herd_task)
        return {'return_value': rval}

    @classmethod
    def move_file_to_s3(cls, fname, bucket_name, remote_name, **botokw):
        """Move data in file descriptor to s3.

        :param fname:        Path to file to move.

        :param bucket_name:  String name of S3 bucket. If this starts with
                             '@' then we write to location bucket[1:] (this
                             is useful for testing).

        :param **botokw:  Keyword args for boto (e.g., profile, key, etc.).

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Move data in file descriptor to s3.

        """
        if not bucket_name:
            raise ValueError('Invalid bucket_name: "%s"' % str(bucket_name))
        if bucket_name[0] == '@':
            logging.info('Using local file location for bucket "%s"',
                         bucket_name)
            remote_name = os.path.join(bucket_name[1:], remote_name)
            os.makedirs(os.path.dirname(remote_name))
            shutil.copy(fname, remote_name)
        else:
            session = boto3.Session(**botokw)
            s3_client = session.client('s3')

            s3_client.upload_file(fname, bucket_name, remote_name)

    @classmethod
    def make_dump_cmdline(cls, ox_herd_task, outfile):
        """Make command line to use to dump database.
        """
        return ['pg_dump', '-w', '-f', outfile, '--dbname=%s' % (
            cls.get_conn_string(ox_herd_task))]

    @classmethod
    def backup_pion_db(cls, ox_herd_task):
        """Do main work of backing up database to S3.
        """
        msgs, status = [], None
        remote_name = 'postgres_backups/%s/%s' % (
            ox_herd_task.prefix, datetime.datetime.utcnow().strftime(
                'backup_%A.sql.gz'))

        with tempfile.TemporaryDirectory() as tmpdir:
            outfile = os.path.join(tmpdir, 'dump.sql')
            msgs.append(cls._do_dump(ox_herd_task, outfile))

            with open(outfile, 'rb') as f_in:
                with gzip.open(outfile + '.gz', 'wb') as f_out:
                    shutil.copyfileobj(f_in, f_out)

            cls.move_file_to_s3(outfile + '.gz', ox_herd_task.bucket_name,
                                remote_name)

        msgs += ['Finished backup succesfully\nStatus=%s\n.' % (
            status)]

        return '\n'.join(msgs)

    @classmethod
    def _do_dump(cls, ox_herd_task, outfile):
        """

        :param ox_herd_task:   Task controlling the dump.

        :param outfile:        String path to output dump.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:   String describing result of dump.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:   Create a subprocess to dump to outfile.

        """
        cmd = cls.make_dump_cmdline(ox_herd_task, outfile)
        logging.info('Running cmd: %s', str(cmd))
        logging.info('Dumping DB to temp file %s and then aws',
                     outfile)
        popen = subprocess.Popen(cmd, stderr=subprocess.PIPE)
        try:
            status = popen.wait(timeout=ox_herd_task.timeout)
        except subprocess.TimeoutExpired as prob:
            logging.error('Timeout while trying to do backup: %s',
                          str(prob))
            raise
        if status != 0:
            msg = 'Got non-zero exist status %s; stderr=%s' % (
                status, popen.stderr.read())
            logging.error(msg)
            raise ValueError(msg)
        return 'Finished dump: extra messages="%s".' % (
            popen.stderr.read())



"""Classes to help automate web tasks.
"""

import re
import logging
import urllib.parse

import requests

from ox_herd.core.plugins import base
from ox_herd.core import ox_tasks


class SimpleTaskResult:
    """Class to hold task result.

Clairfying the expected return values for a task makes it easier to display
and inspect task results with other tools.
    """

    def __init__(self, return_value: str, full_text: str = None,
                 status_code: int = 0, reason: str = 'OK',
                 extras: dict = None):
        """Initializer.

        :param return_value:str: String return value to display as task
                                 status.

        :param full_text:str:   Full text of result/response. This is typically
                                the full response to an HTTP request. You can
                                provide '' if you want to save space. If you
                                provide None, then we use return_value.

        :param status_code:int: Status code from an HTTP request. Use 0 if
                                was not an HTTP request.

        :param reason:str='OK': String reason provided for HTTP response. Use
                                'NA' if not an HTTP response.

        :param extras=None:  Optional dict of additional data.

        """
        self.return_value = return_value
        self.full_text = full_text if full_text is not None else return_value
        self.status_code = status_code
        self.reason = reason
        self.extras = extras if extras else {}

    @classmethod
    def fields(cls) -> list:
        """Return list of strings describing main fields in self.

Sub-classes can override if they want additional fields to showup
in to_dict.
        """

        _ = cls
        return ['return_value', 'full_text', 'status_code',
                'reason', 'extras']

    def to_dict(self) -> dict:
        "Return dict with data in self.fields()"
        return {n: getattr(self, n) for n in self.fields()}


class SimpleWebTask(ox_tasks.OxHerdTask, base.OxPluginComponent):
    """Generic command that can be sub-classed for automation.
    """

    def __init__(self, *args, base_url=None, base_port='', **kwargs):
        """Initializer.

        :arg base_url=None:        String url for where your server lives (or
                                   where the version you want to use lives).
        """
        if not args and 'name' not in kwargs:
            kwargs['name'] = self.__class__.__name__
        ox_tasks.OxHerdTask.__init__(self, *args, **kwargs)

        # Never use base_url/base_port directly. Instead call
        # set_base_url and make_url so we handle ports correctly.
        self.__base_url = ''
        self.__base_port = None
        self.set_base_url(base_url, base_port)

    def set_base_url(self, base_url: str, base_port: str = None):
        """Set the base_url (and optinally base_port).

        :param base_url:    String like http://foo or http://bar:9999

        :param base_port=None: Optional base_port if not in base_url.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Set base URL and port to use in connections.

        """
        self.__base_url = base_url if base_url else self.make_base_url()
        if not self.__base_url:
            return
        my_base, my_port = urllib.parse.splitport(self.__base_url)
        if my_port:
            self.__base_url = my_base  # take off port and store in base_port
            if base_port:
                raise ValueError('base_url=%s has port and base_port=%s' % (
                    self.__base_url, base_port))
            base_port = my_port
        self.__base_port = base_port

    def make_url(self, path: str = '') -> str:
        """Make url with given path.

        :param path='': Optional relative path.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:  Full string URL.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Combine base URL and port set with set_base_url with path.

        """
        result = self.__base_url
        if not result:
            logging.error('self.__dict__: %s', str(self.__dict__))
            raise ValueError('Must have base_url set to call make_url')
        if self.__base_port:
            result += ':' + str(self.__base_port)
        result += '/' + path.lstrip('/')

        return result

    @classmethod
    def make_base_url(cls) -> str:
        """Return string indicating base URL or None if not known.

Sub-classes can override if they know a good default. Otherwise,
should call set_base_url when they have the url to use.
        """
        return None

    def get_login_info(self) -> (str, str):
        """Return string login field and login value

Sub-classes may want to override.
        """
        return 'username', getattr(self, 'username', 'test_user')

    def get_password(self, login_name: str) -> str:
        "Get password for given login name"
        return self.get_secret(login_name, 'test_passwords')

    @classmethod
    def get_login_route(cls) -> str:
        "Return path to login route."

        _ = cls
        return '/login'

    @classmethod
    def get_secret(cls, name: str, category: str = 'root') -> str:
        """Lookup a secret like a password or something.

        :param name:     Name of secret to lookup

        :param category='root':    Category of secret.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:  A string reprsenting the secret you want to get.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Provide a way to lookup secrets like passwords, etc.

        """
        raise NotImplementedError

    def setup_session(self):
        """Setup a requests.session and return it.

This sets up a session so we are logged in to whatever we get
from self.make_url(self.get_login_route()) using the login from
get_login_info().
        """
        session = requests.session()
        my_url = self.make_url(self.get_login_route())
        csrf_field, csrf = None, None
        try:
            csrf_field, csrf = self.get_csrf_from_form(session, my_url)
        except Exception as problem:
            logging.error('Failed to get csrf due to exception: %s', str(
                problem))
            logging.error('Maybe override get_csrf_from_form?')
            raise

        login_field, login_name = self.get_login_info()
        password = self.get_password(login_name)
        data = {login_field: login_name, 'password': password}
        if csrf:
            data[csrf_field] = csrf
        post_resp = session.post(my_url, data, verify=False)
        if post_resp.status_code != 200:
            raise ValueError('Got unexpected status/reason: %s/%s in login' % (
                post_resp.status_code, post_resp.reason))
        return session

    def do_main(self, session) -> SimpleTaskResult:
        """Sub-classes should override this to do the main work.

Sub-classes should override this to return an instance of SimpleTaskResult
describing the result of running the task. See docs for SimpleTaskResult
for more details.
        """
        raise NotImplementedError

    @classmethod
    def main_call(cls, ox_herd_task):
        logging.info('Starting main_call for %s', cls.__name__)
        session = ox_herd_task.setup_session()
        result = ox_herd_task.do_main(session)
        msg = 'Go return_value %s' % (result.return_value)
        cls.note_comment(ox_herd_task, msg)

        return result

    @classmethod
    def note_comment(cls, ox_herd_task: ox_tasks.OxHerdTask,
                     comment: str):
        """Note a comment for a task we ran.

        :param ox_herd_task:    Task we ran.

        :param comment:         Comment to note.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Mainly a placeholder in case sub-classes want
                  to do some kind of logging for comments.

        """
        _ = cls
        logging.info('Comment for task %s: %s', ox_herd_task.name,
                     comment)

    @staticmethod
    def get_csrf_from_form(
            session, url: str, csrf_field: str = 'csrf_token',
            csrf_re: str = None) -> (str, str):
        """Do a get request for the given url and extract CSRF token.

        :param session:    Session we have to the web site.

        :param url:        String URL for form with CSRF token.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:  String name for CSRF field and string for CSRF token.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Flask WTForms provides a CSRF token to prevent CSRF
                  attacks. We need to hit the url with a GET request to
                  get the csrf_token and include that as a parameter of
                  our POST request. This function gets the csrf_token.
        """
        result = session.get(url)
        csrf_re = csrf_re if csrf_re else ' *'.join([
            'id="%s"' % csrf_field, 'name="%s"' % csrf_field,
            'type="hidden"', 'value="(?P<csrf>[^"]*)">'])
        if isinstance(csrf_re, str):
            csrf_re = re.compile(csrf_re)
        match = csrf_re.search(result.text)
        if not match:
            raise ValueError('Could not extract csrf from url "%s"' % url)
        return csrf_field, match.group('csrf')

    def raise_on_bad_status(self, result):
        """Raise ValueError if http response looks like an error.

        :param result:        Response from an HTTP command such as what
                              do_main might report.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:   False if no problems, otherwise raises an exception.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:   Raise an exception if result looks like an error.

        """
        if result.status_code != 200:
            msg = 'Problem in task %s with reason: "%s"' % (
                self.__class__.__name__, result.reason)
            logging.error(msg)
            raise ValueError(msg)
        if ' error' in result.text.lower():
            self.note_comment(self, 'Saw error in result: ' + str(result))
            if isinstance(result.reason, str):
                my_reason = result.reason
            else:
                my_reason = result.reason.text
            msg = 'Saw "error" in result for task %s: "%s"' % (
                self.__class__.__name__, my_reason)
            logging.error(msg)
            raise ValueError(msg)

        return False



"""Provides pylint plugin for ox_herd.
"""

from flask import Blueprint
from ox_herd.core.plugins.pylint_plugin.core import OxHerdPyLintPlugin

OxHerdPyLintPlugin.set_bp(Blueprint(
    'pylint_plugin', __name__, template_folder='templates'))

def get_ox_plugin():
    """Required function for module to provide plugin.
    """
    return OxHerdPyLintPlugin()





"""Tools to check and verify health.

"""

import doctest
import time
import logging
import datetime
import threading
import typing

import requests

from redis import Redis
from rq import Queue, Worker
from rq_scheduler import Scheduler


class PostSuccess:
    """Class to post to a URL on success.

Can be provided for `success` argument to `ProbeQueue`.

By doing something like

    health.RQDoc.default_success = health.PostSuccess(
            url='https://hc-ping.com/<your-key>/background-tasks',
            headers={'User-Agent': 'ox_herd task monitoring'})

you can make it so that when the ox_herd health check succeeds, it
posts a message to hc-ping and use hc-piong to alert you if ox_herd
stops responding.

    """

    def __init__(self, url: str, **request_kwargs):
        """Initializer.

        :param url:  String URL to POST to.

        :param **request_kwargs:  Keyword arguments for requests.post

        """
        self.url = url
        self.request_kwargs = dict(request_kwargs)

    def __call__(self, msg: str):
        result = requests.post(self.url, **self.request_kwargs)
        logging.info('PostSuccess to %s returned results: %s',
                     self.url, result.text)


class ThresholdComplainer:
    """Optional complaining callable to use in RQDoc.

This class provides a callable which can track complaints over a period of time
and only raise an exception once we have too many compalints in that period.

An instance of this class can be passed as the `complain` argument to RQDoc.
    """

    def __init__(self, raise_threshold: int = 4,
                 raise_lookback_seconds: float = 60*60*8):
        """Initializer.

        :param raise_threshold=4:  How many complaints to tolerate before raising
                                   an exception.

        :param raise_lookback_seconds=60*60*8:  Tracking period. We forget complaints
                                                that happened more than this many
                                                seconds ago.

        """
        self.raise_threshold = raise_threshold
        self.raise_lookback_seconds = raise_lookback_seconds
        self.error_list = []

    def maybe_complain(self, msg):
        """Track the given complaint in `msg`.
        """
        now = datetime.datetime.utcnow()
        self.error_list.append((now, msg))
        keep_idx = 0
        while keep_idx < len(self.error_list) and ((
                now - self.error_list[keep_idx][0]
                ).total_seconds() >= self.raise_lookback_seconds):
            logging.info('Forget about old complaint at index %i: %s',
                         keep_idx, self.error_list[keep_idx])
            keep_idx += 1
        self.error_list = self.error_list[keep_idx:]
        if len(self.error_list) > self.raise_threshold:
            raise ValueError(f'Got {len(self.error_list)} complaints in lookback period'
                             f'; latest: {msg}')
        logging.warning('Tracking complaint %s but not enough to raise an error yet',
                        msg)

    def __call__(self, msg):
        return self.maybe_complain(msg)


class RQDoc:
    """Doctor to check on health of python rq services.

This class is meant to be able to check on python rq services.
The idea is that you can call the check method which does
some things to verify liveness. See docs on that as well
as _regr_test method for details.
    """

    default_complain = ValueError
    default_success = logging.info

    def __init__(self, complain=None, q_mode: str = 's',
                 success=None):
        """Initializer.

        :param complain=None:  Optional callable which takes a single
                               string argument descricing problem we had
                               in probing a queue and complains. See
                               ProbeQueue class for details. This is used
                               as default when we call self.launch_probe.

                               If you provide None, then we use the value
                               of the class variable default_complain. This
                               lets you change the default_complain to affect
                               global default behaviour.

        :param q_mode='s':     How to enqueue the job when checking queues:
                                 - 's':  Use scheduler to enqueue so that
                                         both queue and scheduler are checked.
                                 - 'q':  Use queue directly so scheduler
                                         not checked.

        :param success=None:   Optional callable which takes a single
                               string argument describing successful
                               result in probing a queue. See
                               ProbeQueue class for details. This is used
                               as default when we call self.launch_probe.

                               If you provide None, then we use the value
                               of the class variable default_success. This
                               lets you change the default_success to affect
                               global default behaviour.

        """
        self.complain = (
            complain if complain else self.__class__.default_complain)
        self.success = (
            success if success else self.__class__.default_success)
        self.q_mode = q_mode

    def check(self, probe_time: int,
              check_queues: typing.Union[str, typing.Sequence[str]],
              complain: callable = None, success=None) -> str:
        """Check liveness.

        :param probe_time:  Integer indicating how long to wait (in seconds)
                            for probing the queue. A value of 0 indicates
                            no probe.

        :param check_queues:  Either a string of the form 'q1/q2' or a sequence
                              of the form ['q1', 'q2'] naming the queues to
                              check for liveness.

        :param complain=None:  Passed to self.launch_probe.

        :param success=None:  Passed to self.launch_probe.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:  'OK' if things look OK.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Calls self.check_workers and self.launch_probe to verify
                  liveness of desired queues. The idea is that this is a
                  high-level method that you can call to do all known checks.

        """
        sdict = {}
        self.check_workers(check_queues)
        if probe_time.strip():
            probe_time = int(probe_time)
            if probe_time < 0:
                raise ValueError('Cannot have negative probe_time')
            for qname in self.queue_name_list(check_queues):
                self.launch_probe(probe_time, qname, sdict, complain,
                                  success=success)
        return 'OK'

    @staticmethod
    def queue_name_list(str_or_seq: typing.Union[str, typing.Sequence[str]]):
        """

        :param str_or_seq:    Either a string of the form 'q1/q2' or a sequence
                              of the form ['q1', 'q2'] naming the queues to
                              check for liveness.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:  List of strings indicating queues to check.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Translate str or sequence into list of strings suitable
                  for other methods.
        """
        if isinstance(str_or_seq, str):
            return str_or_seq.split('/')
        return str_or_seq

    def check_workers(
            self, check_queues: typing.Union[str, typing.Sequence[str]]):
        """Check that workers are alive.

        :param check_queues:  List or sequence of strings indicating queues
                              to check. See queue_name_list for format if
                              you want to pass a string instead of list.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:  The string 'OK' if workers are alive or raises ValueError
                  if something is wrong.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Go through queues in check_queues and see if there is
                  a worker in python rq to work on that queue.

        """
        queue_counts = {}
        worker_list = Worker.all(connection=Redis())
        for worker in worker_list:
            for qname in worker.queue_names():
                queue_counts[qname] = 1 + queue_counts.get(qname, 0)
        for qname in self.queue_name_list(check_queues):
            if not queue_counts.get(qname, None):
                raise ValueError(f'No workers found for queue "{qname}"')
        return 'OK'

    def launch_probe(self, probe_time: int, qname: str, sdict: dict,
                     complain: callable = None, success=None):
        """Launch a probe into the given queue to verify things work.

        :param probe_time: Integer probe time > 0.

        :param qname:  String name of queue to check.

        :param sdict:  Dictionary which will have sdict['status'] set to
                       either 'good' or 'bad' depending on outcome of
                       probe.

        :param complain=None:  Optional callable which takes a single
                               string argument descricing problem we had
                               in probing a queue and complains. If this
                               is None, we use self.complain.

        :param success=None:   Optional callable which takes a single
                               string argument describing succesful
                               probe. If this is None, we use self.success.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Create an instance of ProbeQueue and start it in
                  a separate thread to verify that we can launch jobs
                  into the queue and have them run. See ProbeQueue for
                  more details.

        """
        probe = ProbeQueue(probe_time, qname, sdict,
                           complain if complain else self.complain,
                           q_mode=self.q_mode,
                           success=(success if success else self.success))
        probe.daemon = True
        probe.start()

    @staticmethod
    def _regr_test():
        """Illustrate how checking things works.

First setup basic boiler-plate imports

>>> import random, time
>>> from redis import Redis
>>> from rq import SimpleWorker, Queue
>>> from ox_herd.core import health

Next setup the RQDoc to check health along with queues and workers.

>>> doc = health.RQDoc(q_mode='q')  # just check mode 'q' for simplicity
>>> qname = 'test_q_%i' % random.randint(0, 100000)
>>> queue = Queue(connection=Redis(), name=qname)
>>> worker = SimpleWorker([queue], connection=queue.connection)

Note that we have not started worker yet so calling check method
will return an exception for our named queue since it has no
running workers.

>>> try:  # verify that we get exception if worker not running
...     doc.check(1, qname)
... except ValueError as problem:
...     print('got problem: %s' % str(problem)) # doctest: +ELLIPSIS
...
got problem: No workers found for queue "test_q_..."

Now we create and queue the doc.check_workers function into the queue.
This is basically just a hack so that when we run the worker in burst
mode the job will get run while the worker is active. We could also
do a potentially cleaner version of this test using threading or
subprocess to have the worker running in the background while we
check that it is alive, but this is a simpler test.

>>> job = queue.enqueue(doc.check_workers, qname)
>>> sdict = {}
>>> probe = doc.launch_probe(1, qname, sdict)
>>> worker.work(burst=True) # verify that doing check with running worker
True
>>> job.result              # is fine with no exception
'OK'
>>> job.get_status()
'finished'
>>> time.sleep(3)
>>> sdict
{'status': 'good'}

Now we are going to verify that the probe will detect a stuck queue.
We enqueue a job that just sleeps to block the queue and then we launch
a quick probe and verify that things look bad.

>>> sjob = queue.enqueue(time.sleep, 3)
>>> sdict = {}
>>> probe = doc.launch_probe(1, qname, sdict)
>>> worker.work(burst=True) # verify that doing check with running worker
True
>>> time.sleep(5)
>>> sdict
{'status': 'bad'}

        """


def return_true():
    """Return True

Simple function used to check liveness of workers.
    """
    return True


class ProbeQueue(threading.Thread):
    """Sub-class of Thread to probe a queue to see if it is working.

The idea is that you create an instance of this class with args
described in __init__ and then start the thread to check if there is
a live working queue.

This needs to be a thread since if you are using this in a web server,
then you may want your server to return a response and do the probe
in the background as a separate thread which reports complaints
via the `complain` argument passed to __init__ and reports success
via the `success` argument passed to __init__.
    """

    def __init__(self, probe_time: int, qname: str, sdict: dict,
                 complain: callable, q_mode: str, *args, success=None,
                 **kwargs):
        """Initializer.

        :param probe_time: Integer probe time > 0.

        :param qname:  String name of queue to check.

        :param sdict:  Dictionary which will have sdict['status'] set to
                       either 'good' or 'bad' depending on outcome of
                       probe.

        :param complain=None:  callable which takes a single
                               string argument descricing problem we had
                               in probing a queue and complains. If this
                               returns an instance of Exception, then we
                               raise that.

                               This method is helpful since this is
                               a thread and so raising an exception may
                               not be seen by main thread. Hence you can
                               provide a custom complainer.

        :param q_mode='s':     How to enqueue the job when checking queues:
                                 - 's':  Use scheduler to enqueue so that
                                         both queue and scheduler are checked.
                                 - 'q':  Use queue directly so scheduler
                                         not checked.

        :param success=None:   A callable which takes a single
                               string argument describing successful probe.

                               This method can be useful to ping something
                               to indicate success.

        :param *args, **kwargs:   Passed to Thread.__init__.
        """
        self.probe_time = probe_time
        self.qname = qname
        self.sdict = sdict
        self.complain = complain
        self.q_mode = q_mode
        self.success = success
        super().__init__(*args, **kwargs)

    def queue_job(self):
        """Enqueue a job based on self.q_mode and return queued job.

This will enqueue directly if q_mode == 'q' and use a scheduler if
the q_mode == 's'>
        """
        args = [return_true]
        kwargs = {'job_ttl': 10*self.probe_time,
                  'job_result_ttl': 20*self.probe_time}
        if self.q_mode == 'q':  # Need to remove job_ for direct queue
            my_queue = Queue(self.qname, connection=Redis())
            if 'job_ttl' in kwargs:
                kwargs['ttl'] = kwargs.pop('job_ttl')
            if 'job_result_ttl' in kwargs:
                kwargs['result_ttl'] = kwargs.pop('job_result_ttl')
            launcher = my_queue.enqueue
        elif self.q_mode == 's':
            sched = Scheduler(queue_name=self.qname, connection=Redis())
            launcher = sched.enqueue_in
            # be careful as enqueue_in does not accept normal args/kwargs
            # like ttl and result_ttol
            args.insert(0, datetime.timedelta(seconds=1))
        else:
            raise ValueError('Invalid q_mode: "%s"' % self.q_mode)
        job = launcher(*args, **kwargs)
        return job

    def run(self):
        """Run the thread.
        """
        start = datetime.datetime.utcnow()
        job = self.queue_job()
        for keep_trying in ([min(self.probe_time, 15),  # do few checks after
                             min(self.probe_time, 30),  # short waits
                             min(self.probe_time, 90)] + [  # then wait
                                 self.probe_time + 1] + [  # given probe time
                    0]):  #  the zero casues us to stop trying
            logging.info('Sleeping %s to wait for %s', keep_trying, job)
            time.sleep(keep_trying)
            now = datetime.datetime.utcnow()
            status = job.get_status()
            if status == 'finished' or (
                    now - start).total_seconds() > self.probe_time:
                break  # either job done or waited too long already
            if not keep_trying:
                self.issue_complaint(
                    f"Couldn't sleep enough {self.probe_time} for job to end")
        assert status == 'finished' or (
            now - start).total_seconds() > self.probe_time
        status = job.get_status()
        if status != 'finished':
            msg = 'At UTC=%s, job %s launched at %s has status %s' % (
                datetime.datetime.utcnow(), job, start, status)
            msg += ('\n*IMPORTANT*:  this could indicate that either the\n'
                    'rq worker or rqscheduler process is *DOWN*')
            self.sdict['status'] = 'bad'
            self.issue_complaint(msg)
        self.sdict['status'] = 'good'
        logging.info('Job %s completed with status %s', job, status)
        if self.success and self.success is not logging.info:
            self.success(f'Job {job} completed with status {status}')
        else:
            logging.info('No success callback configured for health check')

    def issue_complaint(self, msg: str):
        """Issue a complaint.

        :param msg:    String message we complain about.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:  The result of calling self.complain(msg).

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Call self.complain on given msg to generate a complaint
                  about something wrong with a queue. If the compalint is
                  a sub-class of Exception, we raise it.
        """
        complaint = self.complain(msg)
        if isinstance(complaint, Exception):
            raise complaint
        return complaint


if __name__ == '__main__':
    doctest.testmod()
    print('Finished Tests')



"""Useful decorators and decorator tools.

The `ox_decs` module contains useful decorators and related tools. See
documentation for the following for detailed information:

  - `fix_doc`
  - `withdoc`
  - `composed`

The `fix_doc` function can be used in your own decorators to make sure
your decorator adds information to the original docstring about the
decorator applied. For already existing decorators, you can use
the `withdoc` function to make the existing decorator call `fix_doc` to
update the docstring. Finally, you can use `composed` to combine
multiple decorators in a way that correctly updates the docstring with
information about the decorators whether or not they used `fix_doc`.

See docs for `withdoc` for more information.
"""

import doctest
import time
import logging
import functools


def withdoc(decorator):
    """Decorator to apply a decorator and attach its docs to a function.

    Basically, you can ust put `@withdoc` as a decorator calling your
    original (unmodified) decorator to get preserved docs.

    The following illustrates example usage:

>>> import functools
>>> def withprint(func):
...     "withprint decorates a function to print result after running it."
...     @functools.wraps(func)
...     def decorated(*args, **kwargs):
...         result = func(*args, **kwargs)
...         print('Output of %s is %s' % (func.__name__, result))
...         return result
...     return decorated
...
>>> @withprint
... def add(x, y):
...    "Add inputs x and y together"
...    return x + y
...
>>> print(add.__doc__)  # Note that it does *NOT* mention decoration
Add inputs x and y together
>>> @withdoc(withprint)
... def combine(x, y):
...    "Add inputs x and y together"
...    return x + y
...
>>> print(combine.__doc__)  # Note that it *DOES* mention decoration
Add inputs x and y together
-------
Wrapped by decorator withprint:
withprint decorates a function to print result after running it.

    """
    @functools.wraps(decorator)
    def dec_with_doc(func):
        "Create decorator with doc."
        fix_doc(func, decorator)
        return decorator(func)
    return dec_with_doc


def fix_doc(func, decorator):
    """Fix a decorated function with docs about its decorator

    :param func:        Decorated function.

    :param decorator:   The decorator that decorated func.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:   Modify func.__doc__ to describe the decorator.
               You can use the `fix_doc` function to make sure your
               decorator fixes docs of a decorated function when writing
               your own decorators as illustrated below.

   SEE ALSO:   See also the `withdoc` which can apply this to existing
               decorators.

>>> import functools
>>> def withfun(func):  # Write a decorator to illustrate how to use fix_doc
...     "Print how fun a function was after calling it."
...     @functools.wraps(func)
...     def decorated(*args, **kwargs):
...         "Function with fun"
...         result = func(*args, **kwargs)
...         name = getattr(func, '__name__', '(unknown)')
...         print('Calling %s was fun!' % name)
...         return result
...     fix_doc(decorated, withfun)
...     return decorated
...
>>> @withfun
... def add(x, y):
...     "Add x and y together."
...     return x + y
...
>>> print(add.__doc__)
Add x and y together.
-------
Wrapped by decorator withfun:
Print how fun a function was after calling it.

    """
    if not func.__doc__:
        func.__doc__ = 'Function %s' % func.__name__
    extra = '\n-------\nWrapped by decorator %s' % getattr(
        decorator, '__name__', '(unknown)')
    if decorator.__doc__:
        extra += ':\n%s' % decorator.__doc__
    func.__wrapdoc__ = extra
    func.__doc__ += extra
    return func


def withlog(func, logfunc=logging.info):
    "Decorate function to log output and illustrate using fix_doc."

    @functools.wraps(func)
    def decorated(*args, **kwargs):
        "Function with logging"
        result = func(*args, **kwargs)
        name = getattr(func, '__name__', '(unknown)')
        logfunc('Calling %s gives:\n%s\n', name, result)
        return result
    # To make sure that our decorated function updates docstrings of the
    # decorated function, we call fix_doc below. Note that fix_doc references
    # name of the decorator itself.
    fix_doc(decorated, withlog)

    return decorated


def withtime(func, logfunc=logging.info):
    "Decorate function to show run time after calling."

    @functools.wraps(func)
    def decorated(*args, **kwargs):
        "Function with time"
        start = time.time()
        result = func(*args, **kwargs)
        secs = time.time() - start
        name = getattr(func, '__name__', '(unknown)')
        logfunc('Calling %s took %.3f seconds' % (name, secs))
        return result
    return decorated


def composed(*decs):
    """Decorator to compose other decorators together.

>>> import functools
>>> def withprint(func):
...     "withprint decorates a function to print result after running it."
...     @functools.wraps(func)
...     def decorated(*args, **kwargs):
...         result = func(*args, **kwargs)
...         print('Output of %s is %s' % (func.__name__, result))
...         return result
...     return decorated
...
>>> @composed(withprint, withlog, withtime)
... def combine(x, y):
...    "Add inputs x and y together"
...    return x + y
...
>>> print(combine.__doc__)  # Note that it *DOES* mention decoration
Add inputs x and y together
-------
Wrapped by decorator withtime:
Decorate function to show run time after calling.
-------
Wrapped by decorator withlog:
Decorate function to log output and illustrate using fix_doc.
-------
Wrapped by decorator withprint:
withprint decorates a function to print result after running it.

    """
    def deco(func):
        "Composed decorator (decorator 0 applied first)"
        for dec in reversed(decs):
            orig = func
            func = dec(func)
            if orig.__doc__ == func.__doc__:
                fix_doc(func, dec)
        return func
    return deco


if __name__ == '__main__':
    doctest.testmod()
    print('Finished tests')



"""Module containing some simple sub-classes of OxHerdTask to illustate usage.
"""

import logging
import os
import tempfile
import json
import doctest
import shlex
import urllib
import urllib.parse
import urllib.request
import re

try:
    import pytest
except ImportError as problem:
    logging.error('Unable to import pytest because:%s\n%s.',
                  str(problem), "won't be able to use RunPyTest.")


from ox_herd.core.ox_tasks import OxHerdTask

class RunPyTest(OxHerdTask):

    def __init__(self, *args, url=None, pytest_cmd=None, json_file=None, **kw):
        """Initializer.

        :arg *args:    Argumnets to OxHerdTask.__init__.

        :arg url:      URL representing where to run pytest on.

        :arg pytest_cmd: String with command line arguments for running pytest.

        :arg json_file=None:   Optional path for where to store json_file
                               with test results. Usually better to leave this
                               as None indicating to just use a temp file.
                               Sometimes can be useful for testing.

        :arg **kw:     Keyword arguments to OxHerdTask.__init__.

        """
        OxHerdTask.__init__(self, *args, **kw)
        self.pytest_cmd = pytest_cmd
        self.json_file = json_file
        self.url = url

    @staticmethod
    def get_template_name():
        "Override to use custom template to report py test results."

        return 'py_test_report.html'

    @classmethod
    def main_call(cls, ox_herd_task):
        test_file = ox_herd_task.json_file if ox_herd_task.json_file else (
            tempfile.mktemp(suffix='.json'))
        url, cmd_line = cls.do_test(ox_herd_task, test_file)
        test_data = cls.make_report(ox_herd_task, test_file, url, cmd_line)
        if not ox_herd_task.json_file:
            logging.debug('Removing temporary json report %s', test_file)
            os.remove(test_file) # remove temp file

        rval = 'completed test: ' + ', '.join(['%s=%s' % (name, test_data[
            'summary'].get(
                name, '0' if name == 'failed' else 'unknown')) for name in [
                    'failed', 'passed', 'duration']])
        return {'return_value' : rval, 'json_blob' : test_data}

    @staticmethod
    def do_test(py_test_args, test_file):
        url = urllib.parse.urlparse(py_test_args.url)
        if url.scheme == 'file':
            cmd_line = [url.path, '--json', test_file, '-v']
            pta = py_test_args.pytest_cmd
            if isinstance(pta, str):
                pta = shlex.split(pta)
            cmd_line.extend(pta)
            logging.info('Running pytest with command arguments of: %s',
                         str(cmd_line))
            pytest.main(cmd_line)
            return url, cmd_line
        else:
            raise ValueError('URL scheme of %s not handled yet.' % url.scheme)

    @staticmethod
    def make_report(my_task, test_json, url, cmd_line):
        test_data = json.load(open(test_json))['report']
        test_data['url'] = url
        test_data['cmd_line'] = cmd_line
        test_data['task_name'] = my_task.name

        return test_data

class ScanSite(OxHerdTask):

    def __init__(self, *args, url=None, regexp='.', encoding='utf-8',
                 show_len=60, **kw):
        """Initializer.

        :arg *args:      Argumnets to OxHerdTask.__init__.

        :arg url:        URL representing where to run.

        :arg regexp='.': String regular expression to scan for.

        :arg encoding='utf-8':  Expected encoding of URL data.

        :arg show_len=60:  Maxium length of regexp hit to show in return val.

        :arg **kw:       Keyword arguments to OxHerdTask.__init__.

        """
        OxHerdTask.__init__(self, *args, **kw)
        self.url = url
        self.regexp = regexp
        self.encoding = encoding
        self.show_len = show_len

    @staticmethod
    def main_call(ox_herd_task):
        my_re = re.compile(ox_herd_task.regexp)
        with urllib.request.urlopen(ox_herd_task.url) as response:
            data = response.read()
            clean = data.decode(ox_herd_task.encoding, errors='ignore')
            match = my_re.search(clean)
            if match:
                result = clean[match.start():match.end()]
                msg = 'Task %s found a match (first %i chars shown):\n%s' % (
                    ox_herd_task.name, ox_herd_task.show_len,
                    result[:ox_herd_task.show_len])
                #ox_herd_task.send_message(msg)#FIXME
                return msg

        return 'No match found for task %s' % (ox_herd_task.name)

    @staticmethod
    def _regr_test():
        """Simple test to show how task would work.

The following illustrates a simple test of how this task would work.
This is mainly a test of the ox_run_db and other infrastructure since
it does not directly use python rq.

>>> from ox_herd.core import simple_ox_tasks, ox_run_db
>>> import os, json, tempfile, datetime, time, random, imp, logging
>>> random_key = random.randint(0,10000000) # so tests do not collide
>>> print('Using random_key = %s' % str(random_key)) # doctest: +ELLIPSIS
Using random_key = ...
>>> logging.info('Switching to random test redis prefix to be isolated.')
>>> ox_run_db.ox_settings.REDIS_PREFIX += ('test_%s' % random_key)
>>> task = simple_ox_tasks.ScanSite(
...     'test_scan', url='http://google.com', regexp='[Gg]oogle')
>>> logging.info('Simulate what ox_herd would do in running the task:')
>>> result = task.run_ox_task(task)
>>> print(result) # doctest: +ELLIPSIS
Task test_scan found a match (first 60 chars shown):
Google
>>> my_redis = ox_run_db.redis.StrictRedis()
>>> keys = my_redis.keys(ox_run_db.ox_settings.REDIS_PREFIX+'*')
>>> len(keys)
1
>>> task_json = my_redis.get(keys[0])
>>> info = json.loads(task_json.decode('utf8'))
>>> print('%s : %s : %s' % (
...     info['task_name'], info['task_status'], info['return_value']))
test_scan : finished : Task test_scan found a match (first 60 chars shown):
Google
>>> my_redis.delete(keys[0])
1
>>> my_redis.keys(ox_run_db.ox_settings.REDIS_PREFIX+'*')
[]
        """

if __name__ == '__main__':
    doctest.testmod()
    print('Tests finished.')



"""Forms for ox_herd commands.
"""

from wtforms import StringField
from ox_herd.core.plugins import base

class PylintForm(base.GenericOxForm):
    """Use this form to enter parameters for a new pylint job to schedule.
    """

    url = StringField('url', [], description=(
        'Location for package to analyze. This can be a github URL or a file\n'
        'location.'))




"""Views for ox_herd flask blueprint.
"""

import datetime
import copy
import logging
import os
import collections
from collections import namedtuple
import json

import markdown

from flask import (render_template, redirect, request, Markup, url_for, abort)
from flask_login import login_required, current_user

from ox_herd.ui.flask_web_ui.ox_herd import core
from ox_herd.core import health
from ox_herd.core import scheduling, ox_run_db
from ox_herd import settings
from ox_herd.core.plugins import (
    manager as plugin_manager, base as ox_herd_base)


def d_to_nt(dictionary):
    "Convert dictionaryt to namedtuple"
    return namedtuple('GenericDict', dictionary.keys())(**dictionary)


def delete_old_data(old_data):
    for old_time, old_file in old_data:
        logging.debug('Deleting old file %s.', old_file)
        os.remove(old_file)


@core.ox_herd_route('/')
@login_required
def home():
    'redirect to index'
    return index()


@core.ox_herd_route('/show_index')
@login_required
def show_index():
    'redirect to index'
    return index()


@core.ox_herd_route('/index')
@login_required
def index():
    """Main page for ox_herd.

    Usually will not be true root since blueprint will
    be registered with url_prefix.
    """
    commands = collections.OrderedDict([
        (name, Markup('<A HREF="%s">%s</A>' % (
            url_for('ox_herd.%s' % name), name))) for name in [
                'show_plugins', 'list_tasks', 'show_scheduled',
                'show_task_log', 'cancel_job', 'cleanup_job']])

    return render_template('ox_herd/templates/intro.html', commands=commands)


@core.ox_herd_route('/list_tasks')
@login_required
def list_tasks():
    "Show list of tasks so you can inspect them."

    my_db = ox_run_db.create()
    limit = int(request.args.get('limit', 100))
    start_utc = request.args.get('start_utc', None)
    end_utc = request.args.get('end_utc', None)
    tasks = my_db.get_tasks(start_utc=start_utc, end_utc=end_utc)
    total = len(tasks)
    tasks = my_db.limit_task_count(tasks, limit)
    return render_template('task_list.html', title='Task List',
                           tasks=tasks, total=total, limit=limit)


@core.ox_herd_route('/show_task_log')
@login_required
def show_task_log():
    "Show log of tasks run."

    run_db = ox_run_db.create()
    start_utc = request.args.get('start_utc', None)
    end_utc = request.args.get('end_utc', None)
    limit = int(request.args.get('limit', 100))
    tasks = run_db.get_tasks(start_utc=start_utc, end_utc=end_utc)
    tasks = run_db.limit_task_count(tasks, limit)
    other = []
    task_dict = collections.OrderedDict([('started', []), ('finished', [])])
    for item in reversed(sorted(
            tasks, key=lambda t: (t.task_end_utc, t.task_start_utc))):
        if item.task_status in task_dict:
            task_dict[item.task_status].append(item)
        else:
            other.append(item)
    task_dict['other'] = other

    return render_template(
        'task_log.html', title='Task log', start_utc=start_utc,
        end_utc=end_utc, task_dict=task_dict, limit=limit)


@core.ox_herd_route('/show_task')
@login_required
def show_task():
    "Show information about a task."

    task_id = request.args.get('task_id', None)
    run_db = ox_run_db.create()
    try:
        task_data = run_db.get_task(task_id)
        if not task_data:
            raise KeyError('No task with id %s' % str(task_id))
    except Exception as problem:  # pylint: disable=broad-except
        return render_template(
            'generic_error.html', title='Could not find task',
            commentary='Could not find task with id %s because %s' % (
                task_id, problem))

    template = request.args.get('template', task_data.template)
    template = template if (
        template and template.strip() and template != 'default') else (
            'generic_ox_task_result.html')
    return render_template(template, title='Task Report', task_data=task_data)


@core.ox_herd_route('/show_scheduled')
@login_required
def show_scheduled():
    queue_names = request.args.get('queue_names', settings.QUEUE_NAMES)
    queue_names = list(sorted(queue_names.split()))
    my_jobs = scheduling.OxScheduler.get_scheduled_jobs()
    failed_jobs = scheduling.OxScheduler.get_failed_jobs()
    queued = scheduling.OxScheduler.get_queued_jobs(queue_names)
    return render_template('task_schedule.html', task_schedule=my_jobs,
                           queue_names=queue_names, failed_jobs=failed_jobs,
                           queued=queued)


@core.ox_herd_route('/show_job')
@login_required
def show_job():
    jid = request.args.get('jid', None)
    if not jid:
        return redirect(url_for('ox_herd.index'))
    else:
        my_job = scheduling.OxScheduler.find_job(jid)
        ox_herd_task = getattr(my_job, 'kwargs', {}).get('ox_herd_task', None)
        if ox_herd_task is None:
            raise ValueError(
                'Job for id %s has no ox_herd_task (job is %s)' % (
                    jid,
                    str(my_job)))  # FIXME: should show nice error not raise
        return render_template('job_info.html', item=my_job)


@core.ox_herd_route('/launch_job')
@login_required
def launch_job():
    jid = request.args.get('jid', None)
    if jid:
        new_job = scheduling.OxScheduler.launch_job(jid)
        new_jid = new_job.id
    else:
        new_jid = None

    return render_template('launch_job.html', jid=new_jid)


@core.ox_herd_route('/cancel_job')
@login_required
def cancel_job():
    jid = request.args.get('jid', None)
    if not jid:
        return render_template('cancel_job.html')

    cancel = scheduling.OxScheduler.cancel_job(jid)
    return render_template('cancel_job.html', jid=jid, cancel=cancel)


@core.ox_herd_route('/cleanup_job')
@login_required
def cleanup_job():
    cleanup = None
    jid = request.args.get('jid', None)
    if jid:
        cleanup = scheduling.OxScheduler.cleanup_job(jid)
    return render_template('cleanup_job.html', jid=jid, cleanup=cleanup)


@core.ox_herd_route('/requeue_job')
@login_required
def requeue_job():
    jid = request.args.get('jid', None)
    if jid:
        requeue = scheduling.OxScheduler.requeue_job(jid)
    return render_template('requeue_job.html', jid=jid, requeue=requeue)


@core.ox_herd_route('/show_plugins')
@login_required
def show_plugins():
    actives = plugin_manager.PluginManager.get_active_plugins()
    components = []
    for name, plug in actives.items():
        comp_list = plug.get_components()
        logging.debug('Processing plugin %s.', name)
        urls = [(c.cmd_name(), '%s?plugname=%s&plugcomp=%s' % (
            url_for('ox_herd.use_plugin'), name, c.cmd_name()))
                for c in comp_list]
        components.append((name, urls))
    return render_template('show_plugins.html', components=components)


@core.ox_herd_route('/use_plugin', methods=['GET', 'POST'])
@login_required
def use_plugin():
    plugname = request.args.get('plugname', '').strip()
    plugcomp = request.args.get('plugcomp', '').strip()
    plugdict = plugin_manager.PluginManager.get_active_plugins()
    if plugname not in plugdict:
        return 'FIXME: plugin named %s not found in %s' % (plugname, list(
            plugdict))
    myplug = plugdict[plugname]
    compdict = dict([(c.cmd_name(), c) for c in myplug.get_components()])
    if plugcomp not in compdict:
        return 'FIXME: component named %s not found' % plugcomp
    my_comp = compdict[plugcomp]
    my_form_cls = my_comp.get_flask_form_via_cls()
    my_form = my_form_cls()
    if my_form.validate_on_submit():
        klass = my_comp.get_ox_task_cls()
        info = klass(name=my_form.name.data)
        my_form.populate_obj(info)
        job = scheduling.OxScheduler.add_to_schedule(info, info.manager)
        return redirect('%s?jid=%s' % (url_for('ox_herd.show_job'), job.id))

    template = my_comp.get_flask_form_template()
    intro = Markup(markdown.markdown(my_form.__doc__, extensions=[
        'fenced_code', 'tables']))
    return render_template(
        template, form=my_form, intro=intro, title=(
            'Form for component %s of plugin %s' % (plugcomp, plugname)))


@core.ox_herd_route('/schedule_job', methods=['GET', 'POST'])
@login_required
def schedule_job():
    """Schedule a job and configure its parameters.
    """
    jid = request.args.get('jid', None)
    my_job = scheduling.OxScheduler.find_job(jid)
    my_args = my_job.kwargs.get('ox_herd_task', None)
    if my_args is None:
        raise ValueError("job %s had no kwargs['ox_herd_task']" % str(my_job))
    my_args = copy.deepcopy(my_args)
    my_form = core.make_form_for_task(my_args)

    if my_form.validate_on_submit():
        my_form.populate_obj(my_args)
        job = scheduling.OxScheduler.add_to_schedule(my_args, getattr(
            my_args, 'manager', 'rq'))
        return redirect('%s?jid=%s' % (url_for('ox_herd.show_job'), job.id))

    my_form.name.data += '_copy'  # change name to add _copy if making new job

    return render_template(
        'ox_wtf.html', form=my_form, title='Schedule Test',
        intro=Markup(markdown.markdown(my_form.__doc__, extensions=[
            'fenced_code', 'tables'])))


@core.ox_herd_route('/delete_task_from_db')
@login_required
def delete_task_from_db():
    """Delete a task from the task database.

    This is mainly intended to be used from the task list and not so
    useful directly since it requires the task_id.
    """
    task_id = request.args.get('task_id', None)
    if task_id:
        run_db = ox_run_db.create()
        run_db.delete_task(task_id)
        return render_template('generic_display.html', commentary=(
            'Delete task with id %s from database.' % task_id))
    return render_template('generic_display.html', commentary=(
        'Found no task_id so cannot do anything.'))


@core.ox_herd_route('/health_check', noauth=True)
def health_check():
    """Check if system and rq worker and scheduler are healthy.

You can set the value of OX_RQ_DOC.complain at runtime to be a function
which takes a string describing problems with th worker queues and reports
them (e.g., by doing sentry.capture or your own custom stuff).
    """
    try:  # Use try block so return 500 if see an exception
        token = request.args.get('token', '')
        if token not in settings.HEALTH_CHECK_TOKENS:
            logging.warning('Invalid token "%s" for health check; abort',
                            token)
            abort(403)
        logging.info('Valid token for "%s" for health_check received',
                     settings.HEALTH_CHECK_TOKENS[token])
        probe_time = request.args.get('probe_time', '900').strip()
        check_queues = request.args.get('check_queues', 'default').strip()
        doc = health.RQDoc()
        result = doc.check(probe_time, check_queues)
        return result
    except Exception as problem:  # pylint: disable=broad-except
        logging.error('Problem in health_check: %s', str(problem))
        abort(500)


@core.ox_herd_route('/get_latest/<task_name>')
@login_required
def get_latest(task_name):
    """Get info on latest finished version of task named <task_name>

If successful, you can do something like the folllowing to get
the completion time of the latest task:

   datetime.datetime.strptime(my_request.json()['task_end_utc'],
      '%Y-%m-%d %H:%M:%S.%f')

    """
    task_result = ox_run_db.create().get_latest(task_name)
    if task_result is None:
        result = {}
    else:
        result = {n: getattr(task_result, n, None) for n in [
            'task_id', 'task_name', 'task_start_utc', 'task_end_utc',
            'return_value', 'json_data', 'pickle_data']}
    return json.dumps(result), 200, {'ContentType': 'application/json'}


def _check_health_token():
    token = request.args.get('token', '')
    if token not in settings.HEALTH_CHECK_TOKENS:
        logging.warning('Invalid token "%s" for health check; abort',
                        token)
        return False
    logging.info('Valid token for "%s" for health_check received',
                 settings.HEALTH_CHECK_TOKENS[token])
    return True


@core.ox_herd_route('/record_finished_job', methods=['POST'], noauth=True)
def record_finished_job():
    """REST API endpoint to record that a job finished.

You can send a request with the following:

  - task_id:
    - The id returned by record_started_job if you have called that.
      If you have not called record_started_job and want to record
      both start and finish in one go then omit this and provide
      task_name instead (see below).
  - task_name:
    - If you have not already called record_started_job and just
      want to record the start and finish of the job, then provide
      a string name in task_name and omit task_id.
  - return_value:
    - Return value of running the task.
    """
    if current_user is None and not _check_health_token():
        abort(403)
    my_db = ox_run_db.create()
    try:
        record = json.loads(request.data)
        task_id = record.get('task_id', None)
        if task_id is None:
            task_id = my_db.record_task_start(record['task_name'])
            record.pop('task_name')
            record['task_id'] = task_id
        my_db.record_task_finish(**record)
    except KeyError as problem:
        return json.dumps({
            'result': 'error',
            'error': 'Could not find required value for "%s"' % str(
                problem)}), 400, {'ContentType': 'application/json'}
    return json.dumps({'result': 'success'}), 200, {
        'ContentType': 'application/json'}


@core.ox_herd_route('/check_jobs', noauth=True)
def check_jobs():
    """Check if various jobs have been running.

The following URL parameters can be provided:
  - names:
    - Comma separated list of names of tasks to check.
  - seconds (default 3600)
    - How many seconds are allowed since last completion of a job
      named in the names parameter.

If a job in names has not finished within the given seconds,
then we complain.
    """
    my_db = ox_run_db.create()
    late_jobs = []
    try:  # Use try block so return 500 if see an exception
        if not _check_health_token():
            abort(403)
        seconds = int(request.args.get('seconds', '3600'))
        name_list = request.args.get('names').split(',')
        skip_days = [int(i) for i in request.args.get('skip_days','').split(',')
                     if i]
        my_now = datetime.datetime.utcnow()
        if my_now.weekday() in skip_days:
            logging.info('Skip check since weekday now is %s in skip_days %s',
                         my_now, skip_days)
        else:
            for name in name_list:
                logging.info('Checking task "%s"', name)
                latest = my_db.get_latest(name)
                if not latest:
                    late_jobs.append((name, 'not found', 'N/A'))
                else:
                    task_end_utc = datetime.datetime.strptime(
                        str(latest.task_end_utc), '%Y-%m-%d %H:%M:%S.%f')
                    gap = (my_now - task_end_utc).total_seconds()
                    if gap > seconds:
                        late_jobs.append((name, task_end_utc, gap))
        if late_jobs:
            msg = '\n'.join(['Found late jobs:'] + [
                '%s: finished at %s which is %s > %s seconds late' % (
                    name, task_end_utc, gap, seconds)
                for (name, task_end_utc, gap) in late_jobs])
            logging.error(msg)
            return json.dumps({'result': 'error', 'error': msg}), 412, {
                'ContentType': 'application/json'}
    except Exception as problem:  # pylint: disable=broad-except
        logging.error('Problem in health_check: %s', str(problem))
        abort(500)
    return json.dumps({'result': 'success'}), 200, {
        'ContentType': 'application/json'}


def message():
    "Show message that imported views."

    logging.info('Imported ox_herd views for flask.')



"""Module containing some plugin to run pytest.
"""

import configparser
import logging
import os
import tempfile
import json
import subprocess
import shlex
import urllib
import urllib.parse
import urllib.request
import hmac

import jinja2

import xmltodict
import yaml


from ox_herd import settings as ox_herd_settings
from ox_herd.core.plugins import base
from ox_herd.core.ox_tasks import OxHerdTask
from ox_herd.core.plugins.pytest_plugin import forms
from ox_herd.core.plugins import post_to_github_plugin

class OxHerdPyTestPlugin(base.OxPlugin):
    """Plugin to provide pytest services for ox_herd
    """

    __blueprint = None

    @classmethod
    def set_bp(cls, my_bp):
        cls.__blueprint = my_bp

    def get_flask_blueprint(self):
        """Implement as required by OxPlugin."""

        return self.__class__.get_bp()

    @classmethod
    def get_bp(cls):
        return cls.__blueprint


    def name(self):
        """Implement as required by OxPlugin."""

        return 'pytest_plugin'

    def description(self):
        """Implement as required by OxPlugin."""

        return "Plugin to provide pytest services for ox_herd."

    def get_components(self):
        return [RunPyTest('plugin component')]

class RunPyTest(OxHerdTask, base.OxPluginComponent):

    def __init__(self, *args, url=None, pytest_cmd=None, xml_file=None,
                 github_info=None, **kw):
        """Initializer.

        :arg *args:    Argumnets to OxHerdTask.__init__.

        :arg url:      URL representing where to run pytest on.

        :arg pytest_cmd: String with command line arguments for running pytest.

        :arg xml_file=None:   Optional path for where to store xml_file
                              with test results. Usually better to leave this
                              as None indicating to just use a temp file.
                              Sometimes can be useful for testing.

        :arg github_info=None: Optional json object containing info about
                               github repo and issue to post comment to.

        :arg **kw:     Keyword arguments to OxHerdTask.__init__.

        """
        OxHerdTask.__init__(self, *args, **kw)
        self.pytest_cmd = pytest_cmd
        self.xml_file = xml_file
        self.url = url
        self.github_info = github_info

    @classmethod
    def make_task_from_request(cls, request, pull_url_type='html'):
        """Make an instance of this task from a web request.

        :arg request:    Web request in json format (e.g., from GitHub webhook)

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :returns:       Instance of cls designed to execute a test based
                        on information from webhook.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:   Create instance of cls based on a hit to the ox_herd/pytest
                   endpoint (which must be of type application/json). This
                   does things like validate the HMAC from github, pull out
                   the payload from the request, and configure the task.
                   After that you can call launch_raw_task on the returned
                   value if you have an OxScheduler object.

        """
        payload = json.loads(request.data.decode('utf8'))
        my_pr = payload['pull_request']
        my_conf, dummy_my_sec = cls._get_config_info(my_pr)
        cls._validate_request(request, my_conf['github_secret'])
        sha = my_pr['head']['sha']
        name = 'github_pr_pytest_%s_%s' % (sha[:10], my_pr['updated_at'])
        task = RunPyTest(
            name=name, url=payload['repository']['%s_url' % pull_url_type],
            pytest_cmd='--doctest-modules', timeout=3000, 
            github_info=my_pr)

        return task

    @staticmethod
    def cmd_name():
        """Provide name as required by OxPluginComponent.
        """
        return 'pytest'

    @staticmethod
    def get_template_name():
        "Override to use custom template to report py test results."

        return 'py_test_report.html'

    @classmethod
    def main_call(cls, ox_herd_task):
        test_file = ox_herd_task.xml_file if ox_herd_task.xml_file else (
            tempfile.mktemp(suffix='.xml'))
        # Create a temporary directory with a context manager so that we
        # can safely use it inside the call and be confident it will get
        # cleaned up properly.
        with tempfile.TemporaryDirectory(suffix='.ox_pytest') as my_tmp_dir:
            url, cmd_line = cls.do_test(ox_herd_task, test_file, my_tmp_dir)
            test_data = cls.make_report(ox_herd_task, test_file, url, cmd_line)
            if not ox_herd_task.xml_file:
                logging.debug('Removing temporary xml report %s', test_file)
                os.remove(test_file) # remove temp file

            cls.post_results_to_github(ox_herd_task, test_data)

        rval = test_data['summary']

        return {'return_value' : rval, 'json_blob' : test_data}

    @classmethod
    def do_test(cls, py_test_args, test_file, my_tmp_dir):
        # Will force PYTHONPATH into my_env to ensure we test the
        # right thing
        my_env = os.environ.copy()
        pta = py_test_args.pytest_cmd
        clone_path = None
        if isinstance(pta, str):
            pta = shlex.split(pta)
        pta.append('--boxed')
        url = urllib.parse.urlparse(py_test_args.url)
        if url.scheme == 'file':
            cmd_line = [url.path, '--junitxml', test_file, '-v'] + pta
            my_env['PYTHONPATH'] = url.path
        elif url.scheme == '' and url.path[:15] == 'git@github.com:':
            clone_path = url.path
        elif url.scheme == 'https':
            my_conf, dummy_sec = cls._get_config_info(py_test_args.github_info)
            if 'github_token' in my_conf:
                clone_path = 'https://%s@%s%s' % (my_conf[
                    'github_token'], url.netloc, url.path)
            else:
                clone_path = url.geturl()
        else:
            raise ValueError('URL scheme/path = "%s/%s" not handled yet.' % (
                url.scheme, url.path))
        if clone_path:
            cls.prep_git_clone(py_test_args, clone_path, my_tmp_dir, my_env)
            cmd_line = [my_tmp_dir, '--junitxml', test_file, '-v'] + pta

        logging.info('Running pytest on %s with command arguments of: %s',
                     my_tmp_dir, str(cmd_line))
        subprocess.call(['py.test'] + cmd_line, env=my_env)
        return url, cmd_line

    @classmethod
    def prep_git_clone(cls, py_test_args, clone_path, my_tmp_dir, my_env):
        # If you are using github, then we need gitpython so import it
        # here so non-github users do not need it
        from git import Repo
        if py_test_args.github_info:
            sha = py_test_args.github_info['head']['sha']
            repo_name = py_test_args.github_info['head']['repo']['name']
        else:
            sha, repo_name = None, os.path.split(
                clone_path)[-1].split('.git')[0]

        logging.warning('Preparing to clone %s', clone_path)
        my_repo = Repo.clone_from(clone_path, my_tmp_dir + '/' + repo_name)
        logging.info('Finished cloning %s', clone_path)
        if sha is not None:
            my_repo.git.checkout(py_test_args.github_info['head']['sha'])
        new_repo = os.path.join(my_tmp_dir, repo_name)
        my_env['PYTHONPATH'] = '%s:%s' % (new_repo, my_tmp_dir)
        yaml_file = os.path.join(new_repo, 'ox_herd_test.yaml')
        if os.path.exists(yaml_file):
            yconfig = yaml.safe_load(open(yaml_file).read())
            my_env['PYTHONPATH'] = ':'.join([name for name in yconfig.pop(
                'prepend_pypaths', [])] + [my_env['PYTHONPATH']])
            for gname, gpath in yconfig.pop('git_clones', {}).items():
                Repo.clone_from(gpath, os.path.join(my_tmp_dir, gname))
            if 'PYTEST_ADDOPTS' in yconfig:
                my_env['PYTEST_ADDOPTS'] = yconfig.pop('PYTEST_ADDOPTS')
            if yconfig:
                logging.warning('Unprocessed items in yaml file: %s', str(
                    yconfig))


    @classmethod
    def post_results_to_github(cls, ox_herd_task, test_data):
        """Helper method to post test results to github.

        :arg ox_herd_task:   The Ox Herd task containing data.

        :arg test_data:      A dictionary containing the result of running
                             tests as produced by make_report.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  If he user has his .ox_herd_conf file setup to include
                  a [pytest/DEFAULT] section (or a section like
                  [pytest/owner/repo]) with a github_user and
                  github_token with access to the repo in
                  ox_herd_task.github_info, then we will try
                  to post the results of the test as a comment in
                  a github issue.

                  This is a key feature in using this plugin for
                  continuous integration with github.
        """
        if not ox_herd_task.github_info:
            return
        grepo = ox_herd_task.github_info['head']['repo']['full_name']
        grepo = grepo.strip()
        sha = ox_herd_task.github_info['head']['sha']
        tmsg = 'Testing commit %s' % sha
        my_conf, dummy_sec = cls._get_config_info(ox_herd_task.github_info)
        msg = '%s\n\nTested %s:\n%s\n' % (tmsg, grepo, test_data['summary'])
        failures = int(test_data['testsuite']['@errors']) + int(
            test_data['testsuite']['@failures'])
        if failures:
            msg += '\n\n' + jinja2.Environment(loader=jinja2.FileSystemLoader(
                os.path.dirname(forms.__file__).rstrip(
                    '/') + '/templates/')).get_template(
                        'py_test_failures.html').render(
                            test_list=test_data['tests'])

        if 'github_issue' in my_conf:
            title = my_conf['github_issue']
            number = None
        else:
            title = ox_herd_task.github_info['title']
            number = ox_herd_task.github_info['number']

        full_repo = ox_herd_task.github_info['head']['repo']['full_name']

        cthread = post_to_github_plugin.PostToGitHub.prep_comment_thread(
            title, number, full_repo, my_conf)
        cthread.add_comment(msg, allow_create=True)

    @classmethod
    def _get_config_info(cls, github_info):
        """Get configuration info from OX_HERD_CONF file based on github_info.

        :arg github_info:    Dictionary with data about github repo or None
                             to use section pytest/DEFAULT

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :returns:   A dictionary from configparser.ConfigParser pulled out of
                    OX_HERD_CONF file based on the repo in github.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:    Figure out which repo we are dealing with and extract
                    configuration from OX_HERD_CONF.

        """
        config_file = cls.get_conf_file()
        my_config = configparser.ConfigParser()
        my_config.read(config_file)
        if github_info:
            owner, repo = github_info['head']['repo']['full_name'].split('/')
            section = 'pytest/%s/%s' % (owner, repo)
        else:
            section = None
        if section is not None and section in my_config:
            my_sec = section
        else:
            my_sec = 'pytest/DEFAULT'

        my_data = my_config[my_sec]

        return my_data, my_sec

    @staticmethod
    def get_conf_file():
        "Helper to deduce config file."

        return ox_herd_settings.OX_HERD_CONF


    @staticmethod
    def _validate_request(request, secret):
        """Validate github signature on request.

        :arg request:   Web request of type application/json

        :arg secret:    Secret used for HMAC.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Verify the HMAC or raise ValueError.

        """
        header_signature = request.headers.get('X-Hub-Signature')
        if header_signature is None:
            raise ValueError(
                'No header signature provided to validate request')

        sha_name, signature = header_signature.split('=')
        if sha_name != 'sha1':
            raise ValueError('Header signature "%s" not supported.' % sha_name)

        request_data = request.data
        mac = hmac.new(bytes(secret, 'utf8'), request_data, digestmod='sha1')
        if not str(mac.hexdigest()) == str(signature):
            raise ValueError('Request digest does not match signature %s' % (
                str(signature)))

    @staticmethod
    def make_report(my_task, test_file, url, cmd_line):
        test_data = xmltodict.parse(open(test_file, 'rb').read(),
                                    xml_attribs=True)
        test_data['url'] = url
        test_data['cmd_line'] = cmd_line
        test_data['task_name'] = my_task.name
        summary_fields = ['errors', 'failures', 'skips', 'tests', 'time']
        fields = test_data['testsuite']
        logging.info('Test fields are %s', ', '.join(map(str, fields)))
        test_data['summary'] = 'Test resultls: ' + ', '.join([
            '%s: %s' % (name, test_data['testsuite'].get(
                '@' + name, 'None'))
            for name in summary_fields])
        test_data['tests'] = test_data['testsuite']['testcase']

        return test_data

    def get_ox_task_cls(self):
        return self.__class__

    def get_flask_form_via_cls(cls):        
        result = forms.SchedJobForm
        logging.debug('Providing form %s for cls %s', result, cls)        
        return result

    @classmethod
    def make_push_warn_task(cls, request, warnables=('refs/heads/master',)):
        """Helper to make a task to warn about direct pushes to master.

        :arg request:    The web request from a github webhook.

        :arg warnables=('refs/heads/master',):  Tuple of strings to warn about.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :returns:  An instance of the PostToGitHub class that when run
                   will post a message to github warning about pushing
                   directly to master.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:   Support continuous integration via pull requests by
                   creating a task that will warn about direct pushes to
                   master. This is intended to be called by the pytest
                   route if a push event is seen.

        """
        payload = json.loads(request.data.decode('utf8'))
        cname = payload.get('head_commit', {}).get('committer', {}).get(
            'name', {})
        if cname == 'GitHub':  # This was a pull request merge so return None
            return None
        gh_info = {'head': {'repo': payload['repository']},
                   'title': 'push_warning', 'number': None}
        my_conf, my_sec = cls._get_config_info(gh_info)
        if payload['ref'] not in warnables:
            logging.debug('Pushing to %s not %s so not warning on push',
                          payload['ref'], str(warnables))
            return None

        full_repo = payload['repository']['full_name']
        title = 'warning_push'
        cls._validate_request(request, my_conf['github_secret'])
        msg = 'Warning: %s pushed to %s on %s' % (
            payload['sender']['login'], payload['ref'], full_repo)

        task = post_to_github_plugin.PostToGitHub(
            msg, full_repo, title, None, cls.get_conf_file(), my_sec,
            name='github_posting')

        return task



"""Provides pytest plugin for ox_herd.
"""

import logging

from flask import Blueprint, request
from ox_herd.core.plugins.pytest_plugin.core import (
    OxHerdPyTestPlugin, RunPyTest)
from ox_herd.core.scheduling import OxScheduler

OH_BP = Blueprint('pytest_plugin', __name__, template_folder='templates')
OxHerdPyTestPlugin.set_bp(OH_BP)


@OH_BP.route('/ox_herd/pytest', defaults={'pull_url_type': 'html'},
             methods=['GET', 'POST'])
@OH_BP.route('/ox_herd/pytest/<pull_url_type>', methods=['GET', 'POST'])
def pytest(pull_url_type='html'):
    """Route for launching pytest directly.

    This route is intended to be called by a github webhook to process
    pull requests. It uses information from the request to create an
    instance of RunPyTest and then launch that job immediately.
    """
    if request.headers['Content-Type'] != 'application/json':
        raise ValueError('Can only process application/json not %s=%s' % (
            'Content-Type', request.headers['Content-Type']))

    event = request.headers['X-Github-Event']
    if event == 'push':
        try:
            warn_task = RunPyTest.make_push_warn_task(request)
            if warn_task is not None:
                warn_job = OxScheduler.launch_raw_task(warn_task)
                msg = 'Launched warn_task %s as job %s' % (warn_task, warn_job)
                logging.debug(msg)
                return msg
        except Exception as prob:
            logging.error('Unable to make warn push task because %s', prob)
            raise

    if event != 'pull_request':
        # Only process pull_request.
        logging.debug('skipping github event %s', event)
        return 'skipped event %s' % str(event)
    if event == 'issue_comment':
        raise ValueError('Triggering on issue_comment can cause inf loop')

    try:
        task = RunPyTest.make_task_from_request(request, pull_url_type)
    except Exception as problem:
        logging.error('RunPyTest.make_task_from_request exception:\n%s\n',
                      problem)
        raise
    my_job = OxScheduler.launch_raw_task(task)
    msg = 'scheduled task from github %s as job %s' % (str(task), my_job)
    logging.debug(msg)

    return msg


def get_ox_plugin():
    """Required function for module to provide plugin.
    """
    return OxHerdPyTestPlugin()



"""Script to start the ox_herd server.

This is the main script to run ox_herd and provide a python Flask
based web server to respond to web requests.
"""

import configparser
import argparse
import logging
import os

from passlib.apps import custom_app_context as pwd_context
from flask import Flask, redirect, url_for
from flask_login import LoginManager
from flask_login import UserMixin, login_required

from ox_herd import settings as ox_herd_settings

DEFAULT_PORT = 6617

def prepare_parser(parser):
    "Prepare an arg parser to read command line."

    parser.add_argument('--debug', type=int, default=0, help=(
        'Use 1 for debug mode else 0; cannot have host=0.0.0.0 w/debug'))
    parser.add_argument('--host', default='0.0.0.0', help=(
        'IP address for allowed host (0.0.0.0 for all access).'))
    parser.add_argument('--plugin', action='append', help=(
        'You can provide as many --plugin options as you like. Each\n'
        'must be the path to an ox herd plugin to enable. For example\n'
        'providing "--plugin ox_herd.core.plugins.pytest_plugin" would\n'
        'enable the pytest plugin if it was not already enabled.'))
    parser.add_argument('--port', default=DEFAULT_PORT, help=(
        'IP port to listen on.'))
    parser.add_argument('--health_token', default=None, help=(
        'Health check token to use for /ox_hed/health_check route'))
    parser.add_argument('--base_url', help=(
        'Base URL to use for ox_herd site. This is usually automatically\n'
        'but you can override when testing.'))
    parser.add_argument('--stub_user', help=(
        'Comma separated <user>:<passwd>,<u2>:<p2>) for stub db for tests.'))
    parser.add_argument('--stub_roles', help=(
        'Username and comma separated roles (as <user>:role_1,role_2).'))
    parser.add_argument('--hash_stub', type=int, default=1, help=(
        'If 1, interpret password in stub_user as hash; if 0 as password.'))
    parser.add_argument('--logging', type=int, default=logging.INFO, help=(
        'Python logLevel. Use %i for DEBUG, %i for INFO, etc.' % (
            logging.DEBUG, logging.INFO)))

def run():
    "Main function to run server."

    parser = argparse.ArgumentParser(
        description='Command line tool to run ox_herd server')
    prepare_parser(parser)
    args = parser.parse_args()
    _do_setup(args)
    _serve(args)

def _do_setup(args):
    "Should be called by run() to do basic setup based on args."

    if args.debug and args.host == '0.0.0.0':
        logging.warning('Setting host to 127.0.0.1 since in debug mode')
        args.host = '127.0.0.1'

    logging.getLogger('').setLevel(args.logging)
    logging.info('Set log level to %s', args.logging)
    plugin_list = args.plugin if args.plugin else []
    plug_set = set(plugin_list)
    if len(plug_set) < len(plugin_list):
        raise ValueError('Duplicates in args.plugin = %s' % plugin_list)
    cur_plugs = set(ox_herd_settings.OX_PLUGINS)
    if args.health_token:
        ox_herd_settings.HEALTH_CHECK_TOKENS[
            args.health_token] = 'set from cmd line'
    if args.stub_roles:
        ox_herd_settings.STUB_USER_ROLES.update({
            args.stub_roles.split(':')[0]:
            args.stub_roles.split(':')[1].split(',')})

    for item in plugin_list:
        if item not in cur_plugs:
            logging.info('Adding plugin %s to OX_PLUGINS.', item)
            ox_herd_settings.OX_PLUGINS.append(item)
        else:
            logging.info(
                'Not adding plugin %s to OX_PLUGINS since already there.', item)


def _do_setup_stub_user(args, app):
    from ox_herd.core import login_stub
    app.register_blueprint(login_stub.LOGIN_STUB_BP)
    for group in args.stub_user.split(','):
        username, password = group.split(':')
        orig_pw = password
        if not args.hash_stub:
            password = pwd_context.hash(password)
        print('-->%s:%s' % ({username: password}, orig_pw))#FIXME
        ox_herd_settings.STUB_USER_DB.update({username: password})


def _setup_stub_login(app, args=None):
    conf_file = ox_herd_settings.OX_HERD_CONF
    if os.path.exists(conf_file):
        from ox_herd.core import login_stub
        app.register_blueprint(login_stub.LOGIN_STUB_BP)
        my_config = configparser.ConfigParser()
        my_config.read(conf_file)
        if 'STUB_USER_DB' in my_config:
            for user, hash_password in my_config.items('STUB_USER_DB'):
                ox_herd_settings.STUB_USER_DB[user] = hash_password
        else:
            logging.warning('Unable to find STUB_USER_DB in conf %s',
                            conf_file)
    else:
        logging.warning('Unable to find OX_HERD_CONF at %s',
                        ox_herd_settings.OX_HERD_CONF)

    if args is not None and args.stub_user:
        _do_setup_stub_user(args, app)
        
    logging.warning('URL rules:\n%s' % str(list(app.url_map.iter_rules())))


def _serve(args):
    "Run the server. Should only be called by run after doing setup."

    if args.host == '0.0.0.0':
        if args.debug:
            raise TypeError('Cannot have host 0.0.0.0 with debug mode')
    else:
        logging.warning('Host = %s. When host != 0.0.0.0 non-local connections '
                        '\nwill be *IGNORED*. Only use for testing.', args.host)

    app = Flask('ox_herd')

    settings = {'SECRET_KEY' : os.urandom(128),
                'USERNAME' : 'admin', 'DEBUG' : args.debug}

    app.config.from_object(__name__)
    app.config.update(settings)

    from ox_herd.ui.flask_web_ui import ox_herd
    from ox_herd.ui.flask_web_ui.ox_herd import views
    app.register_blueprint(ox_herd.OX_HERD_BP, url_prefix='/ox_herd')
    _setup_stub_login(app, args)

    assert bool(settings['DEBUG']) == bool(args.debug), (
        'Inconsistent debug values from settings and args.')

    @app.route("/")
    def redirect_to_ox_herd():
        """Simple redirect to blueprint root.

        This is required so we can redirect from the top-level path
        when running ox_herd in stand-alone mode.
        """
        return redirect(url_for("ox_herd.index"))
    logging.debug('Created %s for initial redirection', redirect_to_ox_herd)

    app.run(host=args.host, debug=args.debug, port=int(args.port))


if __name__ == '__main__':
    run()



"""Base classes for ox_herd plugins.
"""

from flask_wtf import FlaskForm
from wtforms import (StringField, RadioField, IntegerField, validators)

from ox_herd.core import ox_tasks

class OxPlugin(object):
    """Abstract class for an ox_herd plugin.
    """

    def name(self):
        """Return string name of the plug-in.
        """
        raise NotImplementedError

    def description(self):
        """Return string description of plug-in.
        """
        raise NotImplementedError

    def get_flask_blueprint(self):
        """Return flask blueprint if one exists for the plugin; else None.
        """
        dummy = self # suppress pylint warnings
        return None

    def get_components(self):
        """Return list of OxPluginComponents provided by this OxPlugin
        """
        raise NotImplementedError

class TrivialOxPlugin(OxPlugin):
    """Trivial implementation of an ox_herd plugin.
    """

    def __init__(self, components, name=None, doc=None):
        self._components = components
        self._name = name
        self._doc = doc

    def name(self):
        """Return self._name if set otherwise use name of class.
        """
        return self._name if self._name is not None else self.__class__.__name__

    def description(self):
        """Use self._doc if set otherwise use docstring as description.
        """
        return self._doc if self._doc is not None else self.__doc__

    def get_components(self):
        return self._components

class GenericOxForm(FlaskForm):
    """Use this form to enter parameters for a new job to schedule.
    """

    name = StringField('name', [], default='test_', description=(
        'String name for the job you are going to schedule.'))

    queue_name = StringField('queue_name', [validators.DataRequired()],
                             default='', description=(
        'String name for the job queue that the task will use.\n'
        'Usually this is "default". If you use other names, you should\n'
        'make sure you have the appropriate workers running for that queue.'))

    manager = RadioField(
        'manager', default='rq', choices=[(name, name) for name in [
            'instant', 'rq']], description=(
                'Backend implementation for test:\n\n'
                'rq      : python-rq backend for automated background runs\n'
                'instant : run instantly (useful for testing).'))

    timeout = IntegerField(
        'timeout', [], default=900, description=(
            'Timeout in seconds to allow for tasks.'))

    cron_string = StringField(
        'cron_string', [], default='5 1 * * *', description=(
            'Cron format string for when to schedule the task.\n'
            'For example, "5 1 * * 3" would be every Wednesday at 1:05 am.\n'
            'This is used for --manager choices such as rq which support cron\n'
            'scheduling. NOTE: cron_string should have 5 fields. If you try \n'
            'to use the non-standard extended cron format with 6 fields, you\n'
            'may get unexpected results.'))


class OxPluginComponent(object):

    def cmd_name(self):
        """Return command name to use in running this component.

        Sub-classes should probably override to provide pretty name.
        """
        return self.__class__.__name__

    def get_flask_form(self):
        """Like get_flask_form_via_cls but instance method (DEPRECATED)

Basically, just use get_flask_form_via_cls unless you REALLY think
you need `self`. If you really think you need `self`, then read below
for reasons that you should consider.

This get_flask_form method is just like the get_flask_form_via_cls
method except that this method is DEPRECATED and
get_flask_form_via_cls is preferred.

In general, it is preferred to use get_flask_form_via_cls because the
latter is a class method and can be called more easily since you do not
need an instance. For example, if you want to be able to know what the
parameters of a task are so that you can call it via a different
interface (e.g., the command line) then you really don't want
an INSTANCE of the task but just want to be able to ask the task CLASS
what the possible input parameters are.

By default, get_flask_form just calls get_flask_form_via_cls for
backward compatability.
        """
        return self.get_flask_form_via_cls()

    @classmethod
    def get_flask_form_via_cls(cls):
        """Return sub-classs of GenericOxForm allowing user to configure job.

The get_flask_form_via_cls method will be called in various places when
the user may want to be able to see a form to configure an instance of
the job. The default is to return GenericOxForm, but users can override
to return a subclass of GenericOxForm with their own form fields and
behavior.

It is desirable to return a form CLASS not an INSTANCE here for various
reasons:

  1. It is sometimes cleaner to have the class for pickling.
  2. We may want the class so we can instaniate it and let the form
     do something dynamic at instantiation time.

This method can return None if configuring job via Flask form not allowed.
        """
        return GenericOxForm

    def get_ox_task_cls(self):
        """Return a sub-class of OxHerdTask
        """
        my_class = self.__class__
        if issubclass(my_class, ox_tasks.OxHerdTask):
            return my_class
        else:
            raise ValueError('Do not know how to produce OxHerdTask from %s' % (
                str(self.__class__.__name__)))

    def get_flask_form_template(self):
        """Return string for jinja template to render result of get_flask_form.

        Default is 'ox_wtf.html' but you can override if desired.
        """
        dummy = self # suppress pylint warnings
        return 'ox_wtf.html'

class OxPlugTask(ox_tasks.OxHerdTask, OxPluginComponent):
    """Example class to show how to make the simplest plugin.

    This is meanly meant to serve as an example of a minimal plugin.
    Things to note are:

       1. We inherit from OxHerdTask and implement main_call to indicate
          the task that this plugin component does.
       2. We inherit from OxPluginComponent to indicate this is a plugin
          component.
    """

    pass
